"""
tests.unit.utils.test_dockermod
===============================

Test the funcs in salt.utils.dockermod and salt.utils.dockermod.translate
"""

import copy
import functools
import logging
import os

import salt.config
import salt.loader
import salt.utils.dockermod.translate.container
import salt.utils.dockermod.translate.network
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from salt.utils.dockermod.translate import helpers as translate_helpers
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class Assert:
    def __init__(self, translator):
        self.translator = translator

    def __call__(self, func):
        self.func = func
        return functools.wraps(func)(
            # pylint: disable=unnecessary-lambda
            lambda testcase, *args, **kwargs: self.wrap(testcase, *args, **kwargs)
            # pylint: enable=unnecessary-lambda
        )

    def wrap(self, *args, **kwargs):
        raise NotImplementedError

    def test_stringlist(self, testcase, name):
        alias = self.translator.ALIASES_REVMAP.get(name)
        # Using file paths here because "volumes" must be passed through this
        # set of assertions and it requires absolute paths.
        if salt.utils.platform.is_windows():
            data = [r"c:\foo", r"c:\bar", r"c:\baz"]
        else:
            data = ["/foo", "/bar", "/baz"]
        for item in (name, alias):
            if item is None:
                continue
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: ",".join(data)}
                ),
                testcase.apply_defaults({name: data}),
            )
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: data}),
                testcase.apply_defaults({name: data}),
            )
            if name != "volumes":
                # Test coercing to string
                testcase.assertEqual(
                    salt.utils.dockermod.translate_input(
                        self.translator, **{item: ["one", 2]}
                    ),
                    testcase.apply_defaults({name: ["one", "2"]}),
                )
        if alias is not None:
            # Test collision
            # sorted() used here because we want to confirm that we discard the
            # alias' value and go with the unsorted version.
            test_kwargs = {name: data, alias: sorted(data)}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=True, **test_kwargs
                ),
                testcase.apply_defaults({name: test_kwargs[name]}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=False, **test_kwargs
                )

    def test_key_value(self, testcase, name, delimiter):
        """
        Common logic for key/value pair testing. IP address validation is
        turned off here, and must be done separately in the wrapped function.
        """
        alias = self.translator.ALIASES_REVMAP.get(name)
        expected = {"foo": "bar", "baz": "qux"}
        vals = "foo{0}bar,baz{0}qux".format(delimiter)
        for item in (name, alias):
            if item is None:
                continue
            for val in (vals, vals.split(",")):
                testcase.assertEqual(
                    salt.utils.dockermod.translate_input(
                        self.translator, validate_ip_addrs=False, **{item: val}
                    ),
                    testcase.apply_defaults({name: expected}),
                )
            # Dictionary input
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, validate_ip_addrs=False, **{item: expected}
                ),
                testcase.apply_defaults({name: expected}),
            )
            # "Dictlist" input from states
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    validate_ip_addrs=False,
                    **{item: [{"foo": "bar"}, {"baz": "qux"}]},
                ),
                testcase.apply_defaults({name: expected}),
            )
        if alias is not None:
            # Test collision
            test_kwargs = {name: vals, alias: f"hello{delimiter}world"}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    validate_ip_addrs=False,
                    ignore_collisions=True,
                    **test_kwargs,
                ),
                testcase.apply_defaults({name: expected}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator,
                    validate_ip_addrs=False,
                    ignore_collisions=False,
                    **test_kwargs,
                )


class assert_bool(Assert):
    """
    Test a boolean value
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        for item in (name, alias):
            if item is None:
                continue
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: True}),
                testcase.apply_defaults({name: True}),
            )
            # These two are contrived examples, but they will test bool-ifying
            # a non-bool value to ensure proper input format.
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: "foo"}),
                testcase.apply_defaults({name: True}),
            )
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: 0}),
                testcase.apply_defaults({name: False}),
            )
        if alias is not None:
            # Test collision
            test_kwargs = {name: True, alias: False}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=True, **test_kwargs
                ),
                testcase.apply_defaults({name: test_kwargs[name]}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=False, **test_kwargs
                )
        return self.func(testcase, *args, **kwargs)


class assert_int(Assert):
    """
    Test an integer value
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        for item in (name, alias):
            if item is None:
                continue
            for val in (100, "100"):
                testcase.assertEqual(
                    salt.utils.dockermod.translate_input(
                        self.translator, **{item: val}
                    ),
                    testcase.apply_defaults({name: 100}),
                )
            # Error case: non-numeric value passed
            with testcase.assertRaisesRegex(
                CommandExecutionError, "'foo' is not an integer"
            ):
                salt.utils.dockermod.translate_input(self.translator, **{item: "foo"})
        if alias is not None:
            # Test collision
            test_kwargs = {name: 100, alias: 200}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=True, **test_kwargs
                ),
                testcase.apply_defaults({name: test_kwargs[name]}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=False, **test_kwargs
                )
        return self.func(testcase, *args, **kwargs)


class assert_string(Assert):
    """
    Test that item is a string or is converted to one
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        # Using file paths here because "working_dir" must be passed through
        # this set of assertions and it requires absolute paths.
        if salt.utils.platform.is_windows():
            data = r"c:\foo"
        else:
            data = "/foo"
        for item in (name, alias):
            if item is None:
                continue
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: data}),
                testcase.apply_defaults({name: data}),
            )
            if name != "working_dir":
                # Test coercing to string
                testcase.assertEqual(
                    salt.utils.dockermod.translate_input(
                        self.translator, **{item: 123}
                    ),
                    testcase.apply_defaults({name: "123"}),
                )
        if alias is not None:
            # Test collision
            test_kwargs = {name: data, alias: data}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=True, **test_kwargs
                ),
                testcase.apply_defaults({name: test_kwargs[name]}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=False, **test_kwargs
                )
        return self.func(testcase, *args, **kwargs)


class assert_int_or_string(Assert):
    """
    Test an integer or string value
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        for item in (name, alias):
            if item is None:
                continue
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: 100}),
                testcase.apply_defaults({name: 100}),
            )
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: "100M"}),
                testcase.apply_defaults({name: "100M"}),
            )
        if alias is not None:
            # Test collision
            test_kwargs = {name: 100, alias: "100M"}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=True, **test_kwargs
                ),
                testcase.apply_defaults({name: test_kwargs[name]}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=False, **test_kwargs
                )
        return self.func(testcase, *args, **kwargs)


class assert_stringlist(Assert):
    """
    Test a comma-separated or Python list of strings
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        self.test_stringlist(testcase, name)
        return self.func(testcase, *args, **kwargs)


class assert_dict(Assert):
    """
    Dictionaries should be untouched, dictlists should be repacked and end up
    as a single dictionary.
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        expected = {"foo": "bar", "baz": "qux"}
        for item in (name, alias):
            if item is None:
                continue
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: expected}
                ),
                testcase.apply_defaults({name: expected}),
            )
            # "Dictlist" input from states
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: [{x: y} for x, y in expected.items()]}
                ),
                testcase.apply_defaults({name: expected}),
            )
            # Error case: non-dictionary input
            with testcase.assertRaisesRegex(
                CommandExecutionError, "'foo' is not a dictionary"
            ):
                salt.utils.dockermod.translate_input(self.translator, **{item: "foo"})
        if alias is not None:
            # Test collision
            test_kwargs = {name: "foo", alias: "bar"}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=True, **test_kwargs
                ),
                testcase.apply_defaults({name: test_kwargs[name]}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=False, **test_kwargs
                )
        return self.func(testcase, *args, **kwargs)


class assert_cmd(Assert):
    """
    Test for a string, or a comma-separated or Python list of strings. This is
    different from a stringlist in that we do not do any splitting. This
    decorator is used both by the "command" and "entrypoint" arguments.
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        for item in (name, alias):
            if item is None:
                continue
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: "foo bar"}
                ),
                testcase.apply_defaults({name: "foo bar"}),
            )
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: ["foo", "bar"]}
                ),
                testcase.apply_defaults({name: ["foo", "bar"]}),
            )
            # Test coercing to string
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: 123}),
                testcase.apply_defaults({name: "123"}),
            )
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: ["one", 2]}
                ),
                testcase.apply_defaults({name: ["one", "2"]}),
            )
        if alias is not None:
            # Test collision
            test_kwargs = {name: "foo", alias: "bar"}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=True, **test_kwargs
                ),
                testcase.apply_defaults({name: test_kwargs[name]}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=False, **test_kwargs
                )
        return self.func(testcase, *args, **kwargs)


class assert_key_colon_value(Assert):
    """
    Test a key/value pair with parameters passed as key:value pairs
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        self.test_key_value(testcase, name, ":")
        return self.func(testcase, *args, **kwargs)


class assert_key_equals_value(Assert):
    """
    Test a key/value pair with parameters passed as key=value pairs
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        self.test_key_value(testcase, name, "=")
        if name == "labels":
            self.test_stringlist(testcase, name)
        return self.func(testcase, *args, **kwargs)


class assert_labels(Assert):
    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        labels = ["foo", "bar=baz", {"hello": "world"}]
        expected = {"foo": "", "bar": "baz", "hello": "world"}
        for item in (name, alias):
            if item is None:
                continue

            testcase.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, **{item: labels}),
                testcase.apply_defaults({name: expected}),
            )
            # Error case: Passed a mutli-element dict in dictlist
            bad_labels = copy.deepcopy(labels)
            bad_labels[-1]["bad"] = "input"
            with testcase.assertRaisesRegex(
                CommandExecutionError, r"Invalid label\(s\)"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: bad_labels}
                )
        return self.func(testcase, *args, **kwargs)


class assert_device_rates(Assert):
    """
    Tests for device_{read,write}_{bps,iops}. The bps values have a "Rate"
    value expressed in bytes/kb/mb/gb, while the iops values have a "Rate"
    expressed as a simple integer.
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        for item in (name, alias):
            if item is None:
                continue

            # Error case: Not an absolute path
            path = os.path.join("foo", "bar", "baz")
            with testcase.assertRaisesRegex(
                CommandExecutionError,
                "Path '{}' is not absolute".format(path.replace("\\", "\\\\")),
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: f"{path}:1048576"}
                )

            if name.endswith("_bps"):
                # Both integer bytes and a string providing a shorthand for kb,
                # mb, or gb can be used, so we need to test for both.
                expected = ({}, [])
                vals = "/dev/sda:1048576,/dev/sdb:1048576"
                for val in (vals, vals.split(",")):
                    testcase.assertEqual(
                        salt.utils.dockermod.translate_input(
                            self.translator, **{item: val}
                        ),
                        testcase.apply_defaults(
                            {
                                name: [
                                    {"Path": "/dev/sda", "Rate": 1048576},
                                    {"Path": "/dev/sdb", "Rate": 1048576},
                                ]
                            }
                        ),
                    )

                vals = "/dev/sda:1mb,/dev/sdb:5mb"
                for val in (vals, vals.split(",")):
                    testcase.assertEqual(
                        salt.utils.dockermod.translate_input(
                            self.translator, **{item: val}
                        ),
                        testcase.apply_defaults(
                            {
                                name: [
                                    {"Path": "/dev/sda", "Rate": "1mb"},
                                    {"Path": "/dev/sdb", "Rate": "5mb"},
                                ]
                            }
                        ),
                    )

                if alias is not None:
                    # Test collision
                    test_kwargs = {
                        name: "/dev/sda:1048576,/dev/sdb:1048576",
                        alias: "/dev/sda:1mb,/dev/sdb:5mb",
                    }
                    testcase.assertEqual(
                        salt.utils.dockermod.translate_input(
                            self.translator, ignore_collisions=True, **test_kwargs
                        ),
                        testcase.apply_defaults(
                            {
                                name: [
                                    {"Path": "/dev/sda", "Rate": 1048576},
                                    {"Path": "/dev/sdb", "Rate": 1048576},
                                ]
                            }
                        ),
                    )
                    with testcase.assertRaisesRegex(
                        CommandExecutionError, "is an alias for.+cannot both be used"
                    ):
                        salt.utils.dockermod.translate_input(
                            self.translator, ignore_collisions=False, **test_kwargs
                        )
            else:
                # The "Rate" value must be an integer
                vals = "/dev/sda:1000,/dev/sdb:500"
                for val in (vals, vals.split(",")):
                    testcase.assertEqual(
                        salt.utils.dockermod.translate_input(
                            self.translator, **{item: val}
                        ),
                        testcase.apply_defaults(
                            {
                                name: [
                                    {"Path": "/dev/sda", "Rate": 1000},
                                    {"Path": "/dev/sdb", "Rate": 500},
                                ]
                            }
                        ),
                    )
                # Test non-integer input
                expected = (
                    {},
                    {item: "Rate '5mb' for path '/dev/sdb' is non-numeric"},
                    [],
                )
                vals = "/dev/sda:1000,/dev/sdb:5mb"
                for val in (vals, vals.split(",")):
                    with testcase.assertRaisesRegex(
                        CommandExecutionError,
                        "Rate '5mb' for path '/dev/sdb' is non-numeric",
                    ):
                        salt.utils.dockermod.translate_input(
                            self.translator, **{item: val}
                        )

                if alias is not None:
                    # Test collision
                    test_kwargs = {
                        name: "/dev/sda:1000,/dev/sdb:500",
                        alias: "/dev/sda:888,/dev/sdb:999",
                    }
                    testcase.assertEqual(
                        salt.utils.dockermod.translate_input(
                            self.translator, ignore_collisions=True, **test_kwargs
                        ),
                        testcase.apply_defaults(
                            {
                                name: [
                                    {"Path": "/dev/sda", "Rate": 1000},
                                    {"Path": "/dev/sdb", "Rate": 500},
                                ]
                            }
                        ),
                    )
                    with testcase.assertRaisesRegex(
                        CommandExecutionError, "is an alias for.+cannot both be used"
                    ):
                        salt.utils.dockermod.translate_input(
                            self.translator, ignore_collisions=False, **test_kwargs
                        )
        return self.func(testcase, *args, **kwargs)


class assert_subnet(Assert):
    """
    Test an IPv4 or IPv6 subnet
    """

    def wrap(self, testcase, *args, **kwargs):  # pylint: disable=arguments-differ
        # Strip off the "test_" from the function name
        name = self.func.__name__[5:]
        alias = self.translator.ALIASES_REVMAP.get(name)
        for item in (name, alias):
            if item is None:
                continue
            for val in ("127.0.0.1/32", "::1/128"):
                log.debug("Verifying '%s' is a valid subnet", val)
                testcase.assertEqual(
                    salt.utils.dockermod.translate_input(
                        self.translator, validate_ip_addrs=True, **{item: val}
                    ),
                    testcase.apply_defaults({name: val}),
                )
            # Error case: invalid subnet caught by validation
            for val in (
                "127.0.0.1",
                "999.999.999.999/24",
                "10.0.0.0/33",
                "::1",
                "feaz::1/128",
                "::1/129",
            ):
                log.debug("Verifying '%s' is not a valid subnet", val)
                with testcase.assertRaisesRegex(
                    CommandExecutionError, f"'{val}' is not a valid subnet"
                ):
                    salt.utils.dockermod.translate_input(
                        self.translator, validate_ip_addrs=True, **{item: val}
                    )

            # This is not valid input but it will test whether or not subnet
            # validation happened
            val = "foo"
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, validate_ip_addrs=False, **{item: val}
                ),
                testcase.apply_defaults({name: val}),
            )

        if alias is not None:
            # Test collision
            test_kwargs = {name: "10.0.0.0/24", alias: "192.168.50.128/25"}
            testcase.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=True, **test_kwargs
                ),
                testcase.apply_defaults({name: test_kwargs[name]}),
            )
            with testcase.assertRaisesRegex(
                CommandExecutionError, "is an alias for.+cannot both be used"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, ignore_collisions=False, **test_kwargs
                )
        return self.func(testcase, *args, **kwargs)


class TranslateBase(TestCase):
    maxDiff = None
    translator = None  # Must be overridden in the subclass

    def apply_defaults(self, ret, skip_translate=None):
        if skip_translate is not True:
            defaults = getattr(self.translator, "DEFAULTS", {})
            for key, val in defaults.items():
                if key not in ret:
                    ret[key] = val
        return ret

    @staticmethod
    def normalize_ports(ret):
        """
        When we translate exposed ports, we can end up with a mixture of ints
        (representing TCP ports) and tuples (representing UDP ports). Python 2
        will sort an iterable containing these mixed types, but Python 3 will
        not. This helper is used to munge the ports in the return data so that
        the resulting list is sorted in a way that can reliably be compared to
        the expected results in the test.

        This helper should only be needed for port_bindings and ports.
        """
        if "ports" in ret[0]:
            tcp_ports = []
            udp_ports = []
            for item in ret[0]["ports"]:
                if isinstance(item, int):
                    tcp_ports.append(item)
                else:
                    udp_ports.append(item)
            ret[0]["ports"] = sorted(tcp_ports) + sorted(udp_ports)
        return ret

    def tearDown(self):
        """
        Test skip_translate kwarg
        """
        name = self.id().split(".")[-1][5:]
        # The below is not valid input for the Docker API, but these
        # assertions confirm that we successfully skipped translation.
        for val in (True, name, [name]):
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, skip_translate=val, **{name: "foo"}
                ),
                self.apply_defaults({name: "foo"}, skip_translate=val),
            )


class TranslateContainerInputTestCase(TranslateBase):
    """
    Tests for salt.utils.dockermod.translate_input(), invoked using
    salt.utils.dockermod.translate.container as the translator module.
    """

    translator = salt.utils.dockermod.translate.container

    @staticmethod
    def normalize_ports(ret):
        """
        When we translate exposed ports, we can end up with a mixture of ints
        (representing TCP ports) and tuples (representing UDP ports). Python 2
        will sort an iterable containing these mixed types, but Python 3 will
        not. This helper is used to munge the ports in the return data so that
        the resulting list is sorted in a way that can reliably be compared to
        the expected results in the test.

        This helper should only be needed for port_bindings and ports.
        """
        if "ports" in ret:
            tcp_ports = []
            udp_ports = []
            for item in ret["ports"]:
                if isinstance(item, int):
                    tcp_ports.append(item)
                else:
                    udp_ports.append(item)
            ret["ports"] = sorted(tcp_ports) + sorted(udp_ports)
        return ret

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_auto_remove(self):
        """
        Should be a bool or converted to one
        """

    def test_binds(self):
        """
        Test the "binds" kwarg. Any volumes not defined in the "volumes" kwarg
        should be added to the results.
        """
        self.assertEqual(
            salt.utils.dockermod.translate_input(
                self.translator, binds="/srv/www:/var/www:ro", volumes="/testing"
            ),
            {"binds": ["/srv/www:/var/www:ro"], "volumes": ["/testing", "/var/www"]},
        )
        self.assertEqual(
            salt.utils.dockermod.translate_input(
                self.translator, binds=["/srv/www:/var/www:ro"], volumes="/testing"
            ),
            {"binds": ["/srv/www:/var/www:ro"], "volumes": ["/testing", "/var/www"]},
        )
        self.assertEqual(
            salt.utils.dockermod.translate_input(
                self.translator,
                binds={"/srv/www": {"bind": "/var/www", "mode": "ro"}},
                volumes="/testing",
            ),
            {
                "binds": {"/srv/www": {"bind": "/var/www", "mode": "ro"}},
                "volumes": ["/testing", "/var/www"],
            },
        )

    @assert_int(salt.utils.dockermod.translate.container)
    def test_blkio_weight(self):
        """
        Should be an int or converted to one
        """

    def test_blkio_weight_device(self):
        """
        Should translate a list of PATH:WEIGHT pairs to a list of dictionaries
        with the following format: {'Path': PATH, 'Weight': WEIGHT}
        """
        for val in ("/dev/sda:100,/dev/sdb:200", ["/dev/sda:100", "/dev/sdb:200"]):
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, blkio_weight_device="/dev/sda:100,/dev/sdb:200"
                ),
                {
                    "blkio_weight_device": [
                        {"Path": "/dev/sda", "Weight": 100},
                        {"Path": "/dev/sdb", "Weight": 200},
                    ]
                },
            )

        # Error cases
        with self.assertRaisesRegex(
            CommandExecutionError, r"'foo' contains 1 value\(s\) \(expected 2\)"
        ):
            salt.utils.dockermod.translate_input(
                self.translator, blkio_weight_device="foo"
            )
        with self.assertRaisesRegex(
            CommandExecutionError, r"'foo:bar:baz' contains 3 value\(s\) \(expected 2\)"
        ):
            salt.utils.dockermod.translate_input(
                self.translator, blkio_weight_device="foo:bar:baz"
            )
        with self.assertRaisesRegex(
            CommandExecutionError, r"Weight 'foo' for path '/dev/sdb' is not an integer"
        ):
            salt.utils.dockermod.translate_input(
                self.translator, blkio_weight_device=["/dev/sda:100", "/dev/sdb:foo"]
            )

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_cap_add(self):
        """
        Should be a list of strings or converted to one
        """

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_cap_drop(self):
        """
        Should be a list of strings or converted to one
        """

    @assert_cmd(salt.utils.dockermod.translate.container)
    def test_command(self):
        """
        Can either be a string or a comma-separated or Python list of strings.
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_cpuset_cpus(self):
        """
        Should be a string or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_cpuset_mems(self):
        """
        Should be a string or converted to one
        """

    @assert_int(salt.utils.dockermod.translate.container)
    def test_cpu_group(self):
        """
        Should be an int or converted to one
        """

    @assert_int(salt.utils.dockermod.translate.container)
    def test_cpu_period(self):
        """
        Should be an int or converted to one
        """

    @assert_int(salt.utils.dockermod.translate.container)
    def test_cpu_shares(self):
        """
        Should be an int or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_detach(self):
        """
        Should be a bool or converted to one
        """

    @assert_device_rates(salt.utils.dockermod.translate.container)
    def test_device_read_bps(self):
        """
        CLI input is a list of PATH:RATE pairs, but the API expects a list of
        dictionaries in the format [{'Path': path, 'Rate': rate}]
        """

    @assert_device_rates(salt.utils.dockermod.translate.container)
    def test_device_read_iops(self):
        """
        CLI input is a list of PATH:RATE pairs, but the API expects a list of
        dictionaries in the format [{'Path': path, 'Rate': rate}]
        """

    @assert_device_rates(salt.utils.dockermod.translate.container)
    def test_device_write_bps(self):
        """
        CLI input is a list of PATH:RATE pairs, but the API expects a list of
        dictionaries in the format [{'Path': path, 'Rate': rate}]
        """

    @assert_device_rates(salt.utils.dockermod.translate.container)
    def test_device_write_iops(self):
        """
        CLI input is a list of PATH:RATE pairs, but the API expects a list of
        dictionaries in the format [{'Path': path, 'Rate': rate}]
        """

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_devices(self):
        """
        Should be a list of strings or converted to one
        """

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_dns_opt(self):
        """
        Should be a list of strings or converted to one
        """

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_dns_search(self):
        """
        Should be a list of strings or converted to one
        """

    def test_dns(self):
        """
        While this is a stringlist, it also supports IP address validation, so
        it can't use the test_stringlist decorator because we need to test both
        with and without validation, and it isn't necessary to make all other
        stringlist tests also do that same kind of testing.
        """
        for val in ("8.8.8.8,8.8.4.4", ["8.8.8.8", "8.8.4.4"]):
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    dns=val,
                    validate_ip_addrs=True,
                ),
                {"dns": ["8.8.8.8", "8.8.4.4"]},
            )

        # Error case: invalid IP address caught by validation
        for val in ("8.8.8.888,8.8.4.4", ["8.8.8.888", "8.8.4.4"]):
            with self.assertRaisesRegex(
                CommandExecutionError, r"'8.8.8.888' is not a valid IP address"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator,
                    dns=val,
                    validate_ip_addrs=True,
                )

        # This is not valid input but it will test whether or not IP address
        # validation happened.
        for val in ("foo,bar", ["foo", "bar"]):
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    dns=val,
                    validate_ip_addrs=False,
                ),
                {"dns": ["foo", "bar"]},
            )

    @assert_string(salt.utils.dockermod.translate.container)
    def test_domainname(self):
        """
        Should be a list of strings or converted to one
        """

    @assert_cmd(salt.utils.dockermod.translate.container)
    def test_entrypoint(self):
        """
        Can either be a string or a comma-separated or Python list of strings.
        """

    @assert_key_equals_value(salt.utils.dockermod.translate.container)
    def test_environment(self):
        """
        Can be passed in several formats but must end up as a dictionary
        mapping keys to values
        """

    def test_extra_hosts(self):
        """
        Can be passed as a list of key:value pairs but can't be simply tested
        using @assert_key_colon_value since we need to test both with and without
        IP address validation.
        """
        for val in ("web1:10.9.8.7,web2:10.9.8.8", ["web1:10.9.8.7", "web2:10.9.8.8"]):
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    extra_hosts=val,
                    validate_ip_addrs=True,
                ),
                {"extra_hosts": {"web1": "10.9.8.7", "web2": "10.9.8.8"}},
            )

        # Error case: invalid IP address caught by validation
        for val in (
            "web1:10.9.8.299,web2:10.9.8.8",
            ["web1:10.9.8.299", "web2:10.9.8.8"],
        ):
            with self.assertRaisesRegex(
                CommandExecutionError, r"'10.9.8.299' is not a valid IP address"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator,
                    extra_hosts=val,
                    validate_ip_addrs=True,
                )

        # This is not valid input but it will test whether or not IP address
        # validation happened.
        for val in ("foo:bar,baz:qux", ["foo:bar", "baz:qux"]):
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    extra_hosts=val,
                    validate_ip_addrs=False,
                ),
                {"extra_hosts": {"foo": "bar", "baz": "qux"}},
            )

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_group_add(self):
        """
        Should be a list of strings or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_hostname(self):
        """
        Should be a string or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_ipc_mode(self):
        """
        Should be a string or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_isolation(self):
        """
        Should be a string or converted to one
        """

    @assert_labels(salt.utils.dockermod.translate.container)
    def test_labels(self):
        """
        Can be passed as a list of key=value pairs or a dictionary, and must
        ultimately end up as a dictionary.
        """

    @assert_key_colon_value(salt.utils.dockermod.translate.container)
    def test_links(self):
        """
        Can be passed as a list of key:value pairs or a dictionary, and must
        ultimately end up as a dictionary.
        """

    def test_log_config(self):
        """
        This is a mixture of log_driver and log_opt, which get combined into a
        dictionary.

        log_driver is a simple string, but log_opt can be passed in several
        ways, so we need to test them all.
        """
        expected = (
            {"log_config": {"Type": "foo", "Config": {"foo": "bar", "baz": "qux"}}},
            {},
            [],
        )
        for val in (
            "foo=bar,baz=qux",
            ["foo=bar", "baz=qux"],
            [{"foo": "bar"}, {"baz": "qux"}],
            {"foo": "bar", "baz": "qux"},
        ):
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, log_driver="foo", log_opt="foo=bar,baz=qux"
                ),
                {"log_config": {"Type": "foo", "Config": {"foo": "bar", "baz": "qux"}}},
            )

        # Ensure passing either `log_driver` or `log_opt` alone works
        self.assertEqual(
            salt.utils.dockermod.translate_input(self.translator, log_driver="foo"),
            {"log_config": {"Type": "foo", "Config": {}}},
        )
        self.assertEqual(
            salt.utils.dockermod.translate_input(
                self.translator, log_opt={"foo": "bar", "baz": "qux"}
            ),
            {"log_config": {"Type": "none", "Config": {"foo": "bar", "baz": "qux"}}},
        )

    @assert_key_equals_value(salt.utils.dockermod.translate.container)
    def test_lxc_conf(self):
        """
        Can be passed as a list of key=value pairs or a dictionary, and must
        ultimately end up as a dictionary.
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_mac_address(self):
        """
        Should be a string or converted to one
        """

    @assert_int_or_string(salt.utils.dockermod.translate.container)
    def test_mem_limit(self):
        """
        Should be a string or converted to one
        """

    @assert_int(salt.utils.dockermod.translate.container)
    def test_mem_swappiness(self):
        """
        Should be an int or converted to one
        """

    @assert_int_or_string(salt.utils.dockermod.translate.container)
    def test_memswap_limit(self):
        """
        Should be a string or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_name(self):
        """
        Should be a string or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_network_disabled(self):
        """
        Should be a bool or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_network_mode(self):
        """
        Should be a string or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_oom_kill_disable(self):
        """
        Should be a bool or converted to one
        """

    @assert_int(salt.utils.dockermod.translate.container)
    def test_oom_score_adj(self):
        """
        Should be an int or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_pid_mode(self):
        """
        Should be a string or converted to one
        """

    @assert_int(salt.utils.dockermod.translate.container)
    def test_pids_limit(self):
        """
        Should be an int or converted to one
        """

    def test_port_bindings(self):
        """
        This has several potential formats and can include port ranges. It
        needs its own test.
        """
        # ip:hostPort:containerPort - Bind a specific IP and port on the host
        # to a specific port within the container.
        bindings = (
            "10.1.2.3:8080:80,10.1.2.3:8888:80,10.4.5.6:3333:3333,"
            "10.7.8.9:14505-14506:4505-4506,10.1.2.3:8080:81/udp,"
            "10.1.2.3:8888:81/udp,10.4.5.6:3334:3334/udp,"
            "10.7.8.9:15505-15506:5505-5506/udp"
        )
        for val in (bindings, bindings.split(",")):
            self.assertEqual(
                self.normalize_ports(
                    salt.utils.dockermod.translate_input(
                        self.translator,
                        port_bindings=val,
                    )
                ),
                {
                    "port_bindings": {
                        80: [("10.1.2.3", 8080), ("10.1.2.3", 8888)],
                        3333: ("10.4.5.6", 3333),
                        4505: ("10.7.8.9", 14505),
                        4506: ("10.7.8.9", 14506),
                        "81/udp": [("10.1.2.3", 8080), ("10.1.2.3", 8888)],
                        "3334/udp": ("10.4.5.6", 3334),
                        "5505/udp": ("10.7.8.9", 15505),
                        "5506/udp": ("10.7.8.9", 15506),
                    },
                    "ports": [
                        80,
                        3333,
                        4505,
                        4506,
                        (81, "udp"),
                        (3334, "udp"),
                        (5505, "udp"),
                        (5506, "udp"),
                    ],
                },
            )

        # ip::containerPort - Bind a specific IP and an ephemeral port to a
        # specific port within the container.
        bindings = (
            "10.1.2.3::80,10.1.2.3::80,10.4.5.6::3333,10.7.8.9::4505-4506,"
            "10.1.2.3::81/udp,10.1.2.3::81/udp,10.4.5.6::3334/udp,"
            "10.7.8.9::5505-5506/udp"
        )
        for val in (bindings, bindings.split(",")):
            self.assertEqual(
                self.normalize_ports(
                    salt.utils.dockermod.translate_input(
                        self.translator,
                        port_bindings=val,
                    )
                ),
                {
                    "port_bindings": {
                        80: [("10.1.2.3",), ("10.1.2.3",)],
                        3333: ("10.4.5.6",),
                        4505: ("10.7.8.9",),
                        4506: ("10.7.8.9",),
                        "81/udp": [("10.1.2.3",), ("10.1.2.3",)],
                        "3334/udp": ("10.4.5.6",),
                        "5505/udp": ("10.7.8.9",),
                        "5506/udp": ("10.7.8.9",),
                    },
                    "ports": [
                        80,
                        3333,
                        4505,
                        4506,
                        (81, "udp"),
                        (3334, "udp"),
                        (5505, "udp"),
                        (5506, "udp"),
                    ],
                },
            )

        # hostPort:containerPort - Bind a specific port on all of the host's
        # interfaces to a specific port within the container.
        bindings = (
            "8080:80,8888:80,3333:3333,14505-14506:4505-4506,8080:81/udp,"
            "8888:81/udp,3334:3334/udp,15505-15506:5505-5506/udp"
        )
        for val in (bindings, bindings.split(",")):
            self.assertEqual(
                self.normalize_ports(
                    salt.utils.dockermod.translate_input(
                        self.translator,
                        port_bindings=val,
                    )
                ),
                {
                    "port_bindings": {
                        80: [8080, 8888],
                        3333: 3333,
                        4505: 14505,
                        4506: 14506,
                        "81/udp": [8080, 8888],
                        "3334/udp": 3334,
                        "5505/udp": 15505,
                        "5506/udp": 15506,
                    },
                    "ports": [
                        80,
                        3333,
                        4505,
                        4506,
                        (81, "udp"),
                        (3334, "udp"),
                        (5505, "udp"),
                        (5506, "udp"),
                    ],
                },
            )

        # containerPort - Bind an ephemeral port on all of the host's
        # interfaces to a specific port within the container.
        bindings = "80,3333,4505-4506,81/udp,3334/udp,5505-5506/udp"
        for val in (bindings, bindings.split(",")):
            self.assertEqual(
                self.normalize_ports(
                    salt.utils.dockermod.translate_input(
                        self.translator,
                        port_bindings=val,
                    )
                ),
                {
                    "port_bindings": {
                        80: None,
                        3333: None,
                        4505: None,
                        4506: None,
                        "81/udp": None,
                        "3334/udp": None,
                        "5505/udp": None,
                        "5506/udp": None,
                    },
                    "ports": [
                        80,
                        3333,
                        4505,
                        4506,
                        (81, "udp"),
                        (3334, "udp"),
                        (5505, "udp"),
                        (5506, "udp"),
                    ],
                },
            )

        # Test a mixture of different types of input
        bindings = (
            "10.1.2.3:8080:80,10.4.5.6::3333,14505-14506:4505-4506,"
            "9999-10001,10.1.2.3:8080:81/udp,10.4.5.6::3334/udp,"
            "15505-15506:5505-5506/udp,19999-20001/udp"
        )
        for val in (bindings, bindings.split(",")):
            self.assertEqual(
                self.normalize_ports(
                    salt.utils.dockermod.translate_input(
                        self.translator,
                        port_bindings=val,
                    )
                ),
                {
                    "port_bindings": {
                        80: ("10.1.2.3", 8080),
                        3333: ("10.4.5.6",),
                        4505: 14505,
                        4506: 14506,
                        9999: None,
                        10000: None,
                        10001: None,
                        "81/udp": ("10.1.2.3", 8080),
                        "3334/udp": ("10.4.5.6",),
                        "5505/udp": 15505,
                        "5506/udp": 15506,
                        "19999/udp": None,
                        "20000/udp": None,
                        "20001/udp": None,
                    },
                    "ports": [
                        80,
                        3333,
                        4505,
                        4506,
                        9999,
                        10000,
                        10001,
                        (81, "udp"),
                        (3334, "udp"),
                        (5505, "udp"),
                        (5506, "udp"),
                        (19999, "udp"),
                        (20000, "udp"),
                        (20001, "udp"),
                    ],
                },
            )

        # Error case: too many items (max 3)
        with self.assertRaisesRegex(
            CommandExecutionError,
            r"'10.1.2.3:8080:80:123' is an invalid port binding "
            r"definition \(at most 3 components are allowed, found 4\)",
        ):
            salt.utils.dockermod.translate_input(
                self.translator, port_bindings="10.1.2.3:8080:80:123"
            )

        # Error case: port range start is greater than end
        for val in (
            "10.1.2.3:5555-5554:1111-1112",
            "10.1.2.3:1111-1112:5555-5554",
            "10.1.2.3::5555-5554",
            "5555-5554:1111-1112",
            "1111-1112:5555-5554",
            "5555-5554",
        ):
            with self.assertRaisesRegex(
                CommandExecutionError,
                r"Start of port range \(5555\) cannot be greater than end "
                r"of port range \(5554\)",
            ):
                salt.utils.dockermod.translate_input(
                    self.translator,
                    port_bindings=val,
                )

        # Error case: non-numeric port range
        for val in (
            "10.1.2.3:foo:1111-1112",
            "10.1.2.3:1111-1112:foo",
            "10.1.2.3::foo",
            "foo:1111-1112",
            "1111-1112:foo",
            "foo",
        ):
            with self.assertRaisesRegex(
                CommandExecutionError, "'foo' is non-numeric or an invalid port range"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator,
                    port_bindings=val,
                )

        # Error case: misatched port range
        for val in ("10.1.2.3:1111-1113:1111-1112", "1111-1113:1111-1112"):
            with self.assertRaisesRegex(
                CommandExecutionError,
                r"Host port range \(1111-1113\) does not have the same "
                r"number of ports as the container port range \(1111-1112\)",
            ):
                salt.utils.dockermod.translate_input(self.translator, port_bindings=val)

        for val in ("10.1.2.3:1111-1112:1111-1113", "1111-1112:1111-1113"):
            with self.assertRaisesRegex(
                CommandExecutionError,
                r"Host port range \(1111-1112\) does not have the same "
                r"number of ports as the container port range \(1111-1113\)",
            ):
                salt.utils.dockermod.translate_input(
                    self.translator,
                    port_bindings=val,
                )

        # Error case: empty host port or container port
        with self.assertRaisesRegex(
            CommandExecutionError, "Empty host port in port binding definition ':1111'"
        ):
            salt.utils.dockermod.translate_input(self.translator, port_bindings=":1111")
        with self.assertRaisesRegex(
            CommandExecutionError,
            "Empty container port in port binding definition '1111:'",
        ):
            salt.utils.dockermod.translate_input(self.translator, port_bindings="1111:")
        with self.assertRaisesRegex(
            CommandExecutionError, "Empty port binding definition found"
        ):
            salt.utils.dockermod.translate_input(self.translator, port_bindings="")

    def test_ports(self):
        """
        Ports can be passed as a comma-separated or Python list of port
        numbers, with '/tcp' being optional for TCP ports. They must ultimately
        be a list of port definitions, in which an integer denotes a TCP port,
        and a tuple in the format (port_num, 'udp') denotes a UDP port. Also,
        the port numbers must end up as integers. None of the decorators will
        suffice so this one must be tested specially.
        """
        for val in (
            "1111,2222/tcp,3333/udp,4505-4506",
            [1111, "2222/tcp", "3333/udp", "4505-4506"],
            ["1111", "2222/tcp", "3333/udp", "4505-4506"],
        ):
            self.assertEqual(
                self.normalize_ports(
                    salt.utils.dockermod.translate_input(
                        self.translator,
                        ports=val,
                    )
                ),
                {"ports": [1111, 2222, 4505, 4506, (3333, "udp")]},
            )

        # Error case: non-integer and non/string value
        for val in (1.0, [1.0]):
            with self.assertRaisesRegex(
                CommandExecutionError, "'1.0' is not a valid port definition"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator,
                    ports=val,
                )

        # Error case: port range start is greater than end
        with self.assertRaisesRegex(
            CommandExecutionError,
            r"Start of port range \(5555\) cannot be greater than end of "
            r"port range \(5554\)",
        ):
            salt.utils.dockermod.translate_input(
                self.translator,
                ports="5555-5554",
            )

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_privileged(self):
        """
        Should be a bool or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_publish_all_ports(self):
        """
        Should be a bool or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_read_only(self):
        """
        Should be a bool or converted to one
        """

    def test_restart_policy(self):
        """
        Input is in the format "name[:retry_count]", but the API wants it
        in the format {'Name': name, 'MaximumRetryCount': retry_count}
        """
        name = "restart_policy"
        alias = "restart"
        for item in (name, alias):
            # Test with retry count
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: "on-failure:5"}
                ),
                {name: {"Name": "on-failure", "MaximumRetryCount": 5}},
            )
            # Test without retry count
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: "on-failure"}
                ),
                {name: {"Name": "on-failure", "MaximumRetryCount": 0}},
            )
            # Error case: more than one policy passed
            with self.assertRaisesRegex(
                CommandExecutionError, "Only one policy is permitted"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator, **{item: "on-failure,always"}
                )

        # Test collision
        test_kwargs = {name: "on-failure:5", alias: "always"}
        self.assertEqual(
            salt.utils.dockermod.translate_input(
                self.translator, ignore_collisions=True, **test_kwargs
            ),
            {name: {"Name": "on-failure", "MaximumRetryCount": 5}},
        )
        with self.assertRaisesRegex(
            CommandExecutionError, "'restart' is an alias for 'restart_policy'"
        ):
            salt.utils.dockermod.translate_input(
                self.translator, ignore_collisions=False, **test_kwargs
            )

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_security_opt(self):
        """
        Should be a list of strings or converted to one
        """

    @assert_int_or_string(salt.utils.dockermod.translate.container)
    def test_shm_size(self):
        """
        Should be a string or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_stdin_open(self):
        """
        Should be a bool or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_stop_signal(self):
        """
        Should be a string or converted to one
        """

    @assert_int(salt.utils.dockermod.translate.container)
    def test_stop_timeout(self):
        """
        Should be an int or converted to one
        """

    @assert_key_equals_value(salt.utils.dockermod.translate.container)
    def test_storage_opt(self):
        """
        Can be passed in several formats but must end up as a dictionary
        mapping keys to values
        """

    @assert_key_equals_value(salt.utils.dockermod.translate.container)
    def test_sysctls(self):
        """
        Can be passed in several formats but must end up as a dictionary
        mapping keys to values
        """

    @assert_dict(salt.utils.dockermod.translate.container)
    def test_tmpfs(self):
        """
        Can be passed in several formats but must end up as a dictionary
        mapping keys to values
        """

    @assert_bool(salt.utils.dockermod.translate.container)
    def test_tty(self):
        """
        Should be a bool or converted to one
        """

    def test_ulimits(self):
        """
        Input is in the format "name=soft_limit[:hard_limit]", but the API
        wants it in the format
        {'Name': name, 'Soft': soft_limit, 'Hard': hard_limit}
        """
        # Test with and without hard limit
        ulimits = "nofile=1024:2048,nproc=50"
        for val in (ulimits, ulimits.split(",")):
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    ulimits=val,
                ),
                {
                    "ulimits": [
                        {"Name": "nofile", "Soft": 1024, "Hard": 2048},
                        {"Name": "nproc", "Soft": 50, "Hard": 50},
                    ]
                },
            )

        # Error case: Invalid format
        with self.assertRaisesRegex(
            CommandExecutionError,
            r"Ulimit definition 'nofile:1024:2048' is not in the format "
            r"type=soft_limit\[:hard_limit\]",
        ):
            salt.utils.dockermod.translate_input(
                self.translator, ulimits="nofile:1024:2048"
            )

        # Error case: Invalid format
        with self.assertRaisesRegex(
            CommandExecutionError,
            r"Limit 'nofile=foo:2048' contains non-numeric value\(s\)",
        ):
            salt.utils.dockermod.translate_input(
                self.translator, ulimits="nofile=foo:2048"
            )

    def test_user(self):
        """
        Must be either username (string) or uid (int). An int passed as a
        string (e.g. '0') should be converted to an int.
        """
        # Username passed as string
        self.assertEqual(
            salt.utils.dockermod.translate_input(self.translator, user="foo"),
            {"user": "foo"},
        )
        for val in (0, "0"):
            self.assertEqual(
                salt.utils.dockermod.translate_input(self.translator, user=val),
                {"user": 0},
            )

        # Error case: non string/int passed
        with self.assertRaisesRegex(
            CommandExecutionError, "Value must be a username or uid"
        ):
            salt.utils.dockermod.translate_input(self.translator, user=["foo"])

        # Error case: negative int passed
        with self.assertRaisesRegex(CommandExecutionError, "'-1' is an invalid uid"):
            salt.utils.dockermod.translate_input(self.translator, user=-1)

    @assert_string(salt.utils.dockermod.translate.container)
    def test_userns_mode(self):
        """
        Should be a bool or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_volume_driver(self):
        """
        Should be a bool or converted to one
        """

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_volumes(self):
        """
        Should be a list of absolute paths
        """
        # Error case: Not an absolute path
        path = os.path.join("foo", "bar", "baz")
        with self.assertRaisesRegex(
            CommandExecutionError,
            "'{}' is not an absolute path".format(path.replace("\\", "\\\\")),
        ):
            salt.utils.dockermod.translate_input(self.translator, volumes=path)

    @assert_stringlist(salt.utils.dockermod.translate.container)
    def test_volumes_from(self):
        """
        Should be a list of strings or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.container)
    def test_working_dir(self):
        """
        Should be a single absolute path
        """
        # Error case: Not an absolute path
        path = os.path.join("foo", "bar", "baz")
        with self.assertRaisesRegex(
            CommandExecutionError,
            "'{}' is not an absolute path".format(path.replace("\\", "\\\\")),
        ):
            salt.utils.dockermod.translate_input(self.translator, working_dir=path)


class TranslateNetworkInputTestCase(TranslateBase):
    """
    Tests for salt.utils.dockermod.translate_input(), invoked using
    salt.utils.dockermod.translate.network as the translator module.
    """

    translator = salt.utils.dockermod.translate.network

    ip_addrs = {
        True: ("10.1.2.3", "::1"),
        False: ("FOO", "0.9.800.1000", "feaz::1", "aj01::feac"),
    }

    @assert_string(salt.utils.dockermod.translate.network)
    def test_driver(self):
        """
        Should be a string or converted to one
        """

    @assert_key_equals_value(salt.utils.dockermod.translate.network)
    def test_options(self):
        """
        Can be passed in several formats but must end up as a dictionary
        mapping keys to values
        """

    @assert_dict(salt.utils.dockermod.translate.network)
    def test_ipam(self):
        """
        Must be a dict
        """

    @assert_bool(salt.utils.dockermod.translate.network)
    def test_check_duplicate(self):
        """
        Should be a bool or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.network)
    def test_internal(self):
        """
        Should be a bool or converted to one
        """

    @assert_labels(salt.utils.dockermod.translate.network)
    def test_labels(self):
        """
        Can be passed as a list of key=value pairs or a dictionary, and must
        ultimately end up as a dictionary.
        """

    @assert_bool(salt.utils.dockermod.translate.network)
    def test_enable_ipv6(self):
        """
        Should be a bool or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.network)
    def test_attachable(self):
        """
        Should be a bool or converted to one
        """

    @assert_bool(salt.utils.dockermod.translate.network)
    def test_ingress(self):
        """
        Should be a bool or converted to one
        """

    @assert_string(salt.utils.dockermod.translate.network)
    def test_ipam_driver(self):
        """
        Should be a bool or converted to one
        """

    @assert_key_equals_value(salt.utils.dockermod.translate.network)
    def test_ipam_opts(self):
        """
        Can be passed in several formats but must end up as a dictionary
        mapping keys to values
        """

    def ipam_pools(self):
        """
        Must be a list of dictionaries (not a dictlist)
        """
        good_pool = {
            "subnet": "10.0.0.0/24",
            "iprange": "10.0.0.128/25",
            "gateway": "10.0.0.254",
            "aux_addresses": {
                "foo.bar.tld": "10.0.0.20",
                "hello.world.tld": "10.0.0.21",
            },
        }
        bad_pools = [
            {
                "subnet": "10.0.0.0/33",
                "iprange": "10.0.0.128/25",
                "gateway": "10.0.0.254",
                "aux_addresses": {
                    "foo.bar.tld": "10.0.0.20",
                    "hello.world.tld": "10.0.0.21",
                },
            },
            {
                "subnet": "10.0.0.0/24",
                "iprange": "foo/25",
                "gateway": "10.0.0.254",
                "aux_addresses": {
                    "foo.bar.tld": "10.0.0.20",
                    "hello.world.tld": "10.0.0.21",
                },
            },
            {
                "subnet": "10.0.0.0/24",
                "iprange": "10.0.0.128/25",
                "gateway": "10.0.0.256",
                "aux_addresses": {
                    "foo.bar.tld": "10.0.0.20",
                    "hello.world.tld": "10.0.0.21",
                },
            },
            {
                "subnet": "10.0.0.0/24",
                "iprange": "10.0.0.128/25",
                "gateway": "10.0.0.254",
                "aux_addresses": {
                    "foo.bar.tld": "10.0.0.20",
                    "hello.world.tld": "999.0.0.21",
                },
            },
        ]
        self.assertEqual(
            salt.utils.dockermod.translate_input(
                self.translator,
                ipam_pools=[good_pool],
            ),
            {"ipam_pools": [good_pool]},
        )
        for bad_pool in bad_pools:
            with self.assertRaisesRegex(CommandExecutionError, "not a valid"):
                salt.utils.dockermod.translate_input(
                    self.translator, ipam_pools=[good_pool, bad_pool]
                )

    @assert_subnet(salt.utils.dockermod.translate.network)
    def test_subnet(self):
        """
        Must be an IPv4 or IPv6 subnet
        """

    @assert_subnet(salt.utils.dockermod.translate.network)
    def test_iprange(self):
        """
        Must be an IPv4 or IPv6 subnet
        """

    def test_gateway(self):
        """
        Must be an IPv4 or IPv6 address
        """
        for val in self.ip_addrs[True]:
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    validate_ip_addrs=True,
                    gateway=val,
                ),
                self.apply_defaults({"gateway": val}),
            )

        for val in self.ip_addrs[False]:
            with self.assertRaisesRegex(
                CommandExecutionError, f"'{val}' is not a valid IP address"
            ):
                salt.utils.dockermod.translate_input(
                    self.translator,
                    validate_ip_addrs=True,
                    gateway=val,
                )
            self.assertEqual(
                salt.utils.dockermod.translate_input(
                    self.translator,
                    validate_ip_addrs=False,
                    gateway=val,
                ),
                self.apply_defaults(
                    {"gateway": val if isinstance(val, str) else str(val)}
                ),
            )

    @assert_key_equals_value(salt.utils.dockermod.translate.network)
    def test_aux_addresses(self):
        """
        Must be a mapping of hostnames to IP addresses
        """
        name = "aux_addresses"
        alias = "aux_address"
        for item in (name, alias):
            for val in self.ip_addrs[True]:
                addresses = {"foo.bar.tld": val}
                self.assertEqual(
                    salt.utils.dockermod.translate_input(
                        self.translator, validate_ip_addrs=True, **{item: addresses}
                    ),
                    self.apply_defaults({name: addresses}),
                )

            for val in self.ip_addrs[False]:
                addresses = {"foo.bar.tld": val}
                with self.assertRaisesRegex(
                    CommandExecutionError, f"'{val}' is not a valid IP address"
                ):
                    salt.utils.dockermod.translate_input(
                        self.translator, validate_ip_addrs=True, **{item: addresses}
                    )
                self.assertEqual(
                    salt.utils.dockermod.translate_input(
                        self.translator,
                        validate_ip_addrs=False,
                        aux_addresses=addresses,
                    ),
                    self.apply_defaults({name: addresses}),
                )


class DockerTranslateHelperTestCase(TestCase):
    """
    Tests for a couple helper functions in salt.utils.dockermod.translate
    """

    def test_get_port_def(self):
        """
        Test translation of port definition (1234, '1234/tcp', '1234/udp',
        etc.) into the format which docker-py uses (integer for TCP ports,
        'port_num/udp' for UDP ports).
        """
        # Test TCP port (passed as int, no protocol passed)
        self.assertEqual(translate_helpers.get_port_def(2222), 2222)
        # Test TCP port (passed as str, no protocol passed)
        self.assertEqual(translate_helpers.get_port_def("2222"), 2222)
        # Test TCP port (passed as str, with protocol passed)
        self.assertEqual(translate_helpers.get_port_def("2222", "tcp"), 2222)
        # Test TCP port (proto passed in port_num, with passed proto ignored).
        # This is a contrived example as we would never invoke the function in
        # this way, but it tests that we are taking the port number from the
        # port_num argument and ignoring the passed protocol.
        self.assertEqual(translate_helpers.get_port_def("2222/tcp", "udp"), 2222)

        # Test UDP port (passed as int)
        self.assertEqual(translate_helpers.get_port_def(2222, "udp"), (2222, "udp"))
        # Test UDP port (passed as string)
        self.assertEqual(translate_helpers.get_port_def("2222", "udp"), (2222, "udp"))
        # Test UDP port (proto passed in port_num
        self.assertEqual(translate_helpers.get_port_def("2222/udp"), (2222, "udp"))

    def test_get_port_range(self):
        """
        Test extracting the start and end of a port range from a port range
        expression (e.g. 4505-4506)
        """
        # Passing a single int should return the start and end as the same value
        self.assertEqual(translate_helpers.get_port_range(2222), (2222, 2222))
        # Same as above but with port number passed as a string
        self.assertEqual(translate_helpers.get_port_range("2222"), (2222, 2222))
        # Passing a port range
        self.assertEqual(translate_helpers.get_port_range("2222-2223"), (2222, 2223))
        # Error case: port range start is greater than end
        with self.assertRaisesRegex(
            ValueError,
            r"Start of port range \(2222\) cannot be greater than end of "
            r"port range \(2221\)",
        ):
            translate_helpers.get_port_range("2222-2221")
        # Error case: non-numeric input
        with self.assertRaisesRegex(
            ValueError, "'2222-bar' is non-numeric or an invalid port range"
        ):
            translate_helpers.get_port_range("2222-bar")



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.openstack_config as openstack_config
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class OpenstackConfigTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.openstack_config
    """

    def setup_loader_modules(self):
        patcher = patch("salt.utils.path.which", MagicMock(return_value=True))
        patcher.start()
        self.addCleanup(patcher.stop)
        return {openstack_config: {}}

    # 'set_' function tests: 1

    def test_set(self):
        """
        Test if it set a value in an OpenStack configuration file.
        """
        mock = MagicMock(
            return_value={"retcode": 0, "stderr": "error", "stdout": "salt"}
        )
        with patch.dict(openstack_config.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(
                openstack_config.set_(
                    "/etc/keystone/keys.conf", "sql", "connection", "foo"
                ),
                "salt",
            )

        mock = MagicMock(
            return_value={"retcode": 1, "stderr": "error", "stdout": "salt"}
        )
        with patch.dict(openstack_config.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                openstack_config.set_,
                "/etc/keystone/keystone.conf",
                "sql",
                "connection",
                "foo",
            )

    # 'get' function tests: 1

    def test_get(self):
        """
        Test if it get a value from an OpenStack configuration file.
        """
        mock = MagicMock(
            return_value={"retcode": 0, "stderr": "error", "stdout": "salt"}
        )
        with patch.dict(openstack_config.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(
                openstack_config.get("/etc/keystone/keys.conf", "sql", "connection"),
                "salt",
            )

        mock = MagicMock(
            return_value={"retcode": 1, "stderr": "error", "stdout": "salt"}
        )
        with patch.dict(openstack_config.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                openstack_config.get,
                "/etc/key/keystone.conf",
                "sql",
                "connection",
            )

    # 'delete' function tests: 1

    def test_delete(self):
        """
        Test if it delete a value from an OpenStack configuration file.
        """
        mock = MagicMock(
            return_value={"retcode": 0, "stderr": "error", "stdout": "salt"}
        )
        with patch.dict(openstack_config.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(
                openstack_config.delete("/etc/keystone/keys.conf", "sql", "connection"),
                "salt",
            )

        mock = MagicMock(
            return_value={"retcode": 1, "stderr": "error", "stdout": "salt"}
        )
        with patch.dict(openstack_config.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                openstack_config.delete,
                "/etc/key/keystone.conf",
                "sql",
                "connection",
            )



"""
Tests for the file state
"""

import logging
import os
import pathlib
import re
import stat
import subprocess
import textwrap

import pytest
from pytestshellutils.utils import ports

import salt.utils.files
import salt.utils.path
import salt.utils.platform
from salt.utils.versions import Version
from tests.conftest import FIPS_TESTRUN

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def context():
    return {"two": "two", "ten": 10}


@pytest.fixture
def content():
    numbers_file_contents = textwrap.dedent(
        """\
    one
    two
    three

    1
    2
    3
    """
    )
    math_file_contents = textwrap.dedent(
        """\
    Five plus five is ten

    Four squared is sixteen
    """
    )
    return numbers_file_contents, math_file_contents


@pytest.fixture
def all_patch_file(salt_master):
    contents = """
    diff -ur a/foo/bar/math.txt b/foo/bar/math.txt
    --- a/foo/bar/math.txt	2018-04-09 18:43:52.883205365 -0500
    +++ b/foo/bar/math.txt	2018-04-09 18:44:58.525061654 -0500
    @@ -1,3 +1,3 @@
    -Five plus five is ten
    +5 + 5 = 10

    -Four squared is sixteen
    +4² = 16
    diff -ur a/foo/numbers.txt b/foo/numbers.txt
    --- a/foo/numbers.txt	2018-04-09 18:43:58.014272504 -0500
    +++ b/foo/numbers.txt	2018-04-09 18:44:46.487905044 -0500
    @@ -1,7 +1,7 @@
    -one
    -two
     three
    +two
    +one

    -1
    -2
     3
    +2
    +1
    """
    patch_filename = "all.patch"
    with salt_master.state_tree.base.temp_file(f"patches/{patch_filename}", contents):
        yield f"salt://patches/{patch_filename}"


@pytest.fixture
def numbers_patch_file(salt_master):
    contents = """
    --- a/foo/numbers.txt	2018-04-09 18:43:58.014272504 -0500
    +++ b/foo/numbers.txt	2018-04-09 18:44:46.487905044 -0500
    @@ -1,7 +1,7 @@
    -one
    -two
     three
    +two
    +one

    -1
    -2
     3
    +2
    +1
    """
    patch_filename = "numbers.patch"
    with salt_master.state_tree.base.temp_file(f"patches/{patch_filename}", contents):
        yield f"salt://patches/{patch_filename}"


@pytest.fixture
def math_patch_file(salt_master):
    contents = """
    --- a/foo/bar/math.txt	2018-04-09 18:43:52.883205365 -0500
    +++ b/foo/bar/math.txt	2018-04-09 18:44:58.525061654 -0500
    @@ -1,3 +1,3 @@
    -Five plus five is ten
    +5 + 5 = 10

    -Four squared is sixteen
    +4² = 16
    """
    patch_filename = "math.patch"
    with salt_master.state_tree.base.temp_file(f"patches/{patch_filename}", contents):
        yield f"salt://patches/{patch_filename}"


@pytest.fixture
def numbers_patch_template(salt_master):
    contents = """
    --- a/foo/numbers.txt	2018-04-09 18:43:58.014272504 -0500
    +++ b/foo/numbers.txt	2018-04-09 18:44:46.487905044 -0500
    @@ -1,7 +1,7 @@
    -one
    -two
     three
    +{{ two }}
    +one

    -1
    -2
     3
    +2
    +1
    """
    patch_filename = "numbers.patch.jinja"
    with salt_master.state_tree.base.temp_file(f"patches/{patch_filename}", contents):
        yield f"salt://patches/{patch_filename}"


@pytest.fixture
def all_patch_template(salt_master):
    contents = """
    diff -ur a/foo/bar/math.txt b/foo/bar/math.txt
    --- a/foo/bar/math.txt	2018-04-09 18:43:52.883205365 -0500
    +++ b/foo/bar/math.txt	2018-04-09 18:44:58.525061654 -0500
    @@ -1,3 +1,3 @@
    -Five plus five is ten
    +5 + 5 = {{ ten }}

    -Four squared is sixteen
    +4² = 16
    diff -ur a/foo/numbers.txt b/foo/numbers.txt
    --- a/foo/numbers.txt	2018-04-09 18:43:58.014272504 -0500
    +++ b/foo/numbers.txt	2018-04-09 18:44:46.487905044 -0500
    @@ -1,7 +1,7 @@
    -one
    -two
     three
    +{{ two }}
    +one

    -1
    -2
     3
    +2
    +1
    """
    patch_filename = "all.patch.jinja"
    with salt_master.state_tree.base.temp_file(f"patches/{patch_filename}", contents):
        yield f"salt://patches/{patch_filename}"


@pytest.fixture(scope="module")
def pillar_tree(salt_master, salt_minion, salt_call_cli):
    top_file = """
    base:
      '{}':
        - basic
    """.format(
        salt_minion.id
    )
    basic_pillar_file = """
    monty: python
    companions:
      three:
        - liz
        - jo
        - sarah jane
    """
    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    basic_tempfile = salt_master.pillar_tree.base.temp_file(
        "basic.sls", basic_pillar_file
    )

    try:
        with top_tempfile, basic_tempfile:
            ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
            assert ret.returncode == 0
            assert ret.data is True
            yield
    finally:
        # Refresh pillar again to cleaup the temp pillar
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True


@pytest.fixture(scope="module")
def salt_secondary_master(request, salt_factories):
    #
    # Enable a secondary Salt master so we can disable follow_symlinks
    #
    publish_port = ports.get_unused_localhost_port()
    ret_port = ports.get_unused_localhost_port()

    config_defaults = {
        "open_mode": True,
        "transport": request.config.getoption("--transport"),
    }
    config_overrides = {
        "interface": "127.0.0.1",
        "fileserver_followsymlinks": False,
        "publish_port": publish_port,
        "ret_port": ret_port,
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }

    factory = salt_factories.salt_master_daemon(
        "secondary-master",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture(scope="module")
def salt_secondary_minion(salt_secondary_master):
    #
    # Enable a secondary Salt minion so we can
    # point it at athe secondary Salt master
    #
    config_defaults = {}
    config_overrides = {
        "master": salt_secondary_master.config["interface"],
        "master_port": salt_secondary_master.config["ret_port"],
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }

    factory = salt_secondary_master.salt_minion_daemon(
        "secondary-minion",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture(scope="module")
def salt_cli_secondary_wrapper(salt_secondary_master, salt_secondary_minion):
    def run_command(*command, **kwargs):
        salt_cli = salt_secondary_master.salt_cli()
        return salt_cli.run(*command, minion_tgt=salt_secondary_minion.id, **kwargs)

    return run_command


@pytest.mark.usefixtures("pillar_tree")
def test_contents_pillar_with_pillar_list(salt_master, salt_call_cli, tmp_path):
    """
    This tests for any regressions for this issue:
    https://github.com/saltstack/salt/issues/30934
    """
    target_path = tmp_path / "add-contents-pillar-target.txt"
    sls_name = "file-contents-pillar"
    sls_contents = """
    add_contents_pillar_sls:
      file.managed:
        - name: {}
        - contents_pillar: companions:three
    """.format(
        target_path
    )
    sls_tempfile = salt_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )
    with sls_tempfile:
        ret = salt_call_cli.run("state.sls", sls_name)
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        # Check to make sure the file was created
        assert target_path.is_file()


@pytest.mark.usefixtures("pillar_tree")
def test_managed_file_with_pillar_sls(salt_master, salt_call_cli, tmp_path):
    """
    Test to ensure pillar data in sls file
    is rendered properly and file is created.
    """
    ret = salt_call_cli.run("pillar.get", "monty")
    assert ret.returncode == 0
    assert ret.data

    target_path = tmp_path / f"file-pillar-{ret.data}-target.txt"
    sls_name = "file-pillar-get"
    sls_contents = (
        """
    {%- set filedir = '"""
        + str(tmp_path).replace("\\", "/")
        + """' %}
    {%- set filename = "file-pillar-{}-target.txt".format(salt["pillar.get"]("monty", "")) %}
    create-file:
      file.managed:
        - name: {{ filedir | path_join(filename) }}
    """
    )
    sls_tempfile = salt_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )
    with sls_tempfile:
        ret = salt_call_cli.run("state.sls", sls_name)
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        # Check to make sure the file was created
        assert target_path.is_file()


@pytest.mark.usefixtures("pillar_tree")
def test_issue_50221(
    salt_master,
    salt_call_cli,
    tmp_path,
    salt_minion,
    ext_pillar_file_tree_root_dir,
):
    expected_content = "abc\n\n\n"
    target_path = tmp_path / "issue-50221-target.txt"
    sls_name = "issue-50221"
    sls_contents = """
    {{ pillar["target-path"] }}:
      file.managed:
        - contents_pillar: issue-50221
    """
    sls_tempfile = salt_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )
    issue_50221_ext_pillar_tempfile = pytest.helpers.temp_file(
        "issue-50221",
        expected_content,
        ext_pillar_file_tree_root_dir / "hosts" / salt_minion.id,
    )
    with sls_tempfile, issue_50221_ext_pillar_tempfile:
        ret = salt_call_cli.run("pillar.get", "issue-50221")
        assert ret.returncode == 0
        assert ret.data
        # The type of new line, ie, `\n` vs `\r\n` is not important
        assert ret.data.replace("\r\n", "\n") == expected_content
        ret = salt_call_cli.run(
            "state.apply", sls_name, pillar={"target-path": str(target_path)}
        )
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        # Check to make sure the file was created
        assert target_path.is_file()
        # The type of new line, ie, `\n` vs `\r\n` is not important
        assert target_path.read_text().replace("\r\n", "\n") == expected_content


@pytest.fixture
def _check_min_patch_version(shell):
    min_patch_ver = "2.6"
    ret = shell.run("patch", "--version")
    assert ret.returncode == 0
    version = ret.stdout.splitlines()[0].split()[-1]
    if Version(version) < Version(min_patch_ver):
        pytest.xfail(
            "Minimum version of patch not found, expecting {}, found {}".format(
                min_patch_ver, version
            )
        )


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_single_file(salt_call_cli, salt_master, tmp_path):
    """
    Test file.patch using a patch applied to a single file
    """
    name_file = tmp_path / "name_file.txt"
    source_file = tmp_path / "source_file.patch"
    name_file_contents = """
    salt
    patch
    file
    """
    source_file_contents = """
    1,3c1,5
    < salt
    < patch
    < file
    ---
    > salt
    > will
    > patch
    > this
    > file
    """
    sls_contents = """
    do-patch:
      file.patch:
        - name: {name_file}
        - source: {source_file}
    """.format(
        name_file=name_file, source_file=source_file
    )
    sls_temp = salt_master.state_tree.base.temp_file("test_patch.sls", sls_contents)
    name_temp = pytest.helpers.temp_file("name_file.txt", name_file_contents, tmp_path)
    source_temp = pytest.helpers.temp_file(
        "source_file.patch", source_file_contents, tmp_path
    )

    with sls_temp, name_temp, source_temp:
        # Store the original contents and make sure they change
        ret = salt_call_cli.run("state.apply", "test_patch")
        # Check to make sure the patch was applied okay
        assert ret.returncode == 0

        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch successfully applied"

        # Re-run the state, should succeed and there should be a message about
        # a partially-applied hunk.
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch was already applied"
        assert state_run["changes"] == {}


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_directory(salt_call_cli, content, all_patch_file, salt_master, tmp_path):
    """
    Test file.patch using a patch applied to a directory, with changes
    spanning multiple files.
    """
    # Create a new unpatched set of files
    os.makedirs(tmp_path / "foo" / "bar")
    numbers_file = tmp_path / "foo" / "numbers.txt"
    math_file = tmp_path / "foo" / "bar" / "math.txt"

    sls_contents = """
        do-patch:
          file.patch:
            - name: {base_dir}
            - source: {all_patch}
            - strip: 1
        """.format(
        base_dir=tmp_path, all_patch=all_patch_file
    )

    sls_tempfile = salt_master.state_tree.base.temp_file("test_patch.sls", sls_contents)
    numbers_tempfile = pytest.helpers.temp_file(numbers_file, content[0], tmp_path)
    math_tempfile = pytest.helpers.temp_file(math_file, content[1], tmp_path)

    with sls_tempfile, numbers_tempfile, math_tempfile:
        # Run the state file
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data
        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch successfully applied"

        # Re-run the state, should succeed and there should be a message about
        # a partially-applied hunk.
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch was already applied"
        assert state_run["changes"] == {}


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_strip_parsing(
    salt_call_cli, content, all_patch_file, salt_master, tmp_path
):
    """
    Test that we successfuly parse -p/--strip when included in the options
    """
    # Create a new unpatched set of files
    os.makedirs(tmp_path / "foo" / "bar")
    numbers_file = tmp_path / "foo" / "numbers.txt"
    math_file = tmp_path / "foo" / "bar" / "math.txt"

    sls_contents = """
        do-patch:
          file.patch:
            - name: {base_dir}
            - source: {all_patch}
            - options: "-p1"
        """.format(
        base_dir=tmp_path, all_patch=all_patch_file
    )

    sls_patch_contents = """
        do-patch:
          file.patch:
            - name: {base_dir}
            - source: {all_patch}
            - strip: 1
        """.format(
        base_dir=tmp_path, all_patch=all_patch_file
    )

    sls_tempfile = salt_master.state_tree.base.temp_file("test_patch.sls", sls_contents)
    sls_patch_tempfile = salt_master.state_tree.base.temp_file(
        "test_patch_strip.sls", sls_patch_contents
    )
    numbers_tempfile = pytest.helpers.temp_file(numbers_file, content[0], tmp_path)
    math_tempfile = pytest.helpers.temp_file(math_file, content[1], tmp_path)

    with sls_tempfile, sls_patch_tempfile, numbers_tempfile, math_tempfile:
        # Run the state using -p1
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data

        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch successfully applied"

        # Re-run the state using --strip=1
        ret = salt_call_cli.run("state.apply", "test_patch_strip")
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch was already applied"
        assert state_run["changes"] == {}


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_saltenv(salt_call_cli, content, math_patch_file, salt_master, tmp_path):
    """
    Test that we attempt to download the patch from a non-base saltenv
    """
    # This state will fail because we don't have a patch file in that
    # environment, but that is OK, we just want to test that we're looking
    # in an environment other than base.
    # Create a new unpatched set of files
    os.makedirs(tmp_path / "foo" / "bar")
    math_file = tmp_path / "foo" / "bar" / "math.txt"

    sls_contents = """
        do-patch:
          file.patch:
            - name: {math_file}
            - source: {math_patch}
            - saltenv: "prod"
        """.format(
        math_file=math_file, math_patch=math_patch_file
    )
    sls_tempfile = salt_master.state_tree.base.temp_file("test_patch.sls", sls_contents)
    math_tempfile = pytest.helpers.temp_file(math_file, content[1], tmp_path)

    with sls_tempfile, math_tempfile:
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 1
        assert ret.data

        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is False
        assert (
            state_run["comment"]
            == f"Source file {math_patch_file} not found in saltenv 'prod'"
        )


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_single_file_failure(
    salt_call_cli, content, numbers_patch_file, salt_master, tmp_path
):
    """
    Test file.patch using a patch applied to a single file. This tests a
    failed patch.
    """
    # Create a new unpatched set of files
    os.makedirs(tmp_path / "foo" / "bar")
    numbers_file = tmp_path / "foo" / "numbers.txt"
    math_file = tmp_path / "foo" / "bar" / "math.txt"
    reject_file = tmp_path / "reject.txt"

    sls_patch_contents = """
        do-patch:
          file.patch:
            - name: {numbers_file}
            - source: {numbers_patch}
        """.format(
        numbers_file=numbers_file, numbers_patch=numbers_patch_file
    )
    sls_patch_reject_contents = """
        do-patch:
          file.patch:
            - name: {numbers_file}
            - source: {numbers_patch}
            - reject_file: {reject_file}
            - strip: 1
        """.format(
        numbers_file=numbers_file,
        numbers_patch=numbers_patch_file,
        reject_file=reject_file,
    )

    sls_tempfile = salt_master.state_tree.base.temp_file(
        "test_patch.sls", sls_patch_contents
    )
    sls_reject_tempfile = salt_master.state_tree.base.temp_file(
        "test_patch_reject.sls", sls_patch_reject_contents
    )
    numbers_tempfile = pytest.helpers.temp_file(numbers_file, content[0], tmp_path)
    math_tempfile = pytest.helpers.temp_file(math_file, content[1], tmp_path)
    reject_tempfile = pytest.helpers.temp_file("reject.txt", "", tmp_path)

    with (
        sls_tempfile
    ), sls_reject_tempfile, numbers_tempfile, math_tempfile, reject_tempfile:
        # Empty the file to ensure that the patch doesn't apply cleanly
        with salt.utils.files.fopen(numbers_file, "w"):
            pass

        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 1
        assert ret.data
        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is False
        assert "Patch would not apply cleanly" in state_run["comment"]

        # Test the reject_file option and ensure that the rejects are written
        # to the path specified.
        ret = salt_call_cli.run("state.apply", "test_patch_reject")
        assert ret.returncode == 1
        assert ret.data

        state_run = next(iter(ret.data.values()))
        assert "Patch would not apply cleanly" in state_run["comment"]
        assert (
            re.match(state_run["comment"], f"saving rejects to (file )?{reject_file}")
            is None
        )


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_directory_failure(
    salt_call_cli, content, all_patch_file, salt_master, tmp_path
):
    """
    Test file.patch using a patch applied to a directory, with changes
    spanning multiple files.
    """
    # Create a new unpatched set of files
    os.makedirs(tmp_path / "foo" / "bar")
    numbers_file = tmp_path / "foo" / "numbers.txt"
    math_file = tmp_path / "foo" / "bar" / "math.txt"
    reject_file = tmp_path / "reject.txt"

    sls_patch_contents = """
        do-patch:
          file.patch:
            - name: {base_dir}
            - source: {all_patch}
            - strip: 1
        """.format(
        base_dir=tmp_path, all_patch=all_patch_file
    )
    sls_patch_reject_contents = """
        do-patch:
          file.patch:
            - name: {base_dir}
            - source: {all_patch}
            - reject_file: {reject_file}
            - strip: 1
        """.format(
        base_dir=tmp_path, all_patch=all_patch_file, reject_file=reject_file
    )
    sls_tempfile = salt_master.state_tree.base.temp_file(
        "test_patch.sls", sls_patch_contents
    )
    sls_reject_tempfile = salt_master.state_tree.base.temp_file(
        "test_patch_reject.sls", sls_patch_reject_contents
    )

    numbers_tempfile = pytest.helpers.temp_file(numbers_file, content[0], tmp_path)
    math_tempfile = pytest.helpers.temp_file(math_file, content[1], tmp_path)
    reject_tempfile = pytest.helpers.temp_file("reject.txt", "", tmp_path)

    with (
        sls_tempfile
    ), sls_reject_tempfile, numbers_tempfile, math_tempfile, reject_tempfile:
        # Empty the file to ensure that the patch doesn't apply cleanly
        with salt.utils.files.fopen(math_file, "w"):
            pass

        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 1
        assert ret.data
        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is False
        assert "Patch would not apply cleanly" in state_run["comment"]

        # Test the reject_file option and ensure that the rejects are written
        # to the path specified.
        ret = salt_call_cli.run("state.apply", "test_patch_reject")
        assert ret.returncode == 1
        assert ret.data

        state_run = next(iter(ret.data.values()))
        assert "Patch would not apply cleanly" in state_run["comment"]
        assert (
            re.match(state_run["comment"], f"saving rejects to (file )?{reject_file}")
            is None
        )


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_single_file_template(
    salt_call_cli, context, content, numbers_patch_template, salt_master, tmp_path
):
    """
    Test file.patch using a patch applied to a single file, with jinja
    templating applied to the patch file.
    """
    # Create a new unpatched set of files
    os.makedirs(tmp_path / "foo" / "bar", exist_ok=True)
    numbers_file = tmp_path / "foo" / "numbers.txt"

    sls_contents = """
        do-patch:
          file.patch:
            - name: {numbers_file}
            - source: {numbers_patch_template}
            - template: "jinja"
            - context: {context}
        """.format(
        numbers_file=numbers_file,
        numbers_patch_template=numbers_patch_template,
        context=context,
    )

    sls_tempfile = salt_master.state_tree.base.temp_file("test_patch.sls", sls_contents)
    numbers_tempfile = pytest.helpers.temp_file(numbers_file, content[0], tmp_path)

    with sls_tempfile, numbers_tempfile:
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data

        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch successfully applied"

        # Re-run the state, should succeed and there should be a message about
        # a partially-applied hunk.
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch was already applied"
        assert state_run["changes"] == {}


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_directory_template(
    salt_call_cli, context, content, all_patch_template, salt_master, tmp_path
):
    """
    Test file.patch using a patch applied to a directory, with changes
    spanning multiple files, and with jinja templating applied to the patch
    file.
    """
    if salt.utils.platform.is_windows() and os.environ.get("PYTHONUTF8", "0") == "0":
        pytest.skip("Test will fail if PYTHONUTF8=1 is not set on windows")
    # Create a new unpatched set of files
    os.makedirs(tmp_path / "foo" / "bar", exist_ok=True)
    numbers_file = tmp_path / "foo" / "numbers.txt"
    math_file = tmp_path / "foo" / "bar" / "math.txt"

    sls_contents = """
        do-patch:
          file.patch:
            - name: {base_dir}
            - source: {all_patch_template}
            - template: "jinja"
            - context: {context}
        """.format(
        base_dir=tmp_path, all_patch_template=all_patch_template, context=context
    )

    sls_tempfile = salt_master.state_tree.base.temp_file("test_patch.sls", sls_contents)
    numbers_tempfile = pytest.helpers.temp_file(numbers_file, content[0], tmp_path)
    math_tempfile = pytest.helpers.temp_file(math_file, content[1], tmp_path)

    with sls_tempfile, numbers_tempfile, math_tempfile:
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data

        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch was already applied"

        # Re-run the state, should succeed and there should be a message about
        # a partially-applied hunk.
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch was already applied"
        assert state_run["changes"] == {}


@pytest.mark.skip_unless_on_windows
@pytest.mark.skip_if_binaries_missing("patch")
@pytest.mark.usefixtures("_check_min_patch_version")
def test_patch_test_mode(
    salt_call_cli,
    content,
    numbers_patch_file,
    tmp_path,
    salt_master,
):
    """
    Test file.patch using test=True
    """
    # Create a new unpatched set of files
    os.makedirs(tmp_path / "foo" / "bar")
    numbers_file = tmp_path / "foo" / "numbers.txt"

    sls_patch_contents = """
        do-patch:
          file.patch:
            - name: {numbers_file}
            - source: {numbers_patch}
        """.format(
        numbers_file=numbers_file, numbers_patch=numbers_patch_file
    )

    sls_patch_tempfile = salt_master.state_tree.base.temp_file(
        "test_patch.sls", sls_patch_contents
    )
    numbers_tempfile = pytest.helpers.temp_file(numbers_file, content[0], tmp_path)

    with sls_patch_tempfile, numbers_tempfile:
        # Test application with test=True mode
        ret = salt_call_cli.run("state.apply", "test_patch", test=True)
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is None
        assert state_run["comment"] == "The patch would be applied"

        # Apply the patch for real. We'll then be able to test below that we
        # exit with a True rather than a None result if test=True is used on an
        # already-applied patch.
        ret = salt_call_cli.run("state.apply", "test_patch")
        assert ret.returncode == 0
        assert ret.data

        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch successfully applied"

        # Run again with test=True. Since the pre-check happens before we do
        # the __opts__['test'] check, we should exit with a True result just
        # the same as if we try to run this state on an already-patched file
        # *without* test=True.
        ret = salt_call_cli.run("state.apply", "test_patch", test=True)
        assert ret.returncode == 0
        assert ret.data

        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True
        assert state_run["comment"] == "Patch was already applied"

        # Empty the file to ensure that the patch doesn't apply cleanly
        with salt.utils.files.fopen(numbers_file, "w"):
            pass

        # Run again with test=True. Similar to the above run, we are testing
        # that we return before we reach the __opts__['test'] check. In this
        # case we should return a False result because we should already know
        # by this point that the patch will not apply cleanly.
        ret = salt_call_cli.run("state.apply", "test_patch", test=True)
        assert ret.returncode == 1
        assert ret.data

        # Check to make sure the patch was applied okay
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is False
        assert "Patch would not apply cleanly" in state_run["comment"]


@pytest.mark.usefixtures("pillar_tree")
def test_recurse(
    salt_master,
    salt_call_cli,
    tmp_path,
    salt_minion,
):
    target_path = tmp_path / "test_recurse"

    sls_name = "file_recurse_test"
    sls_contents = """
    tmp_recurse_dir_dir:
      file.recurse:
        - name: {target_path}
        - source: salt://tmp_dir/
        - keep_symlinks: True
    """.format(
        target_path=target_path
    )

    test_tempdir = salt_master.state_tree.base.write_path / "tmp_dir"
    test_tempdir.mkdir(parents=True, exist_ok=True)
    sls_tempfile = salt_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )

    with sls_tempfile:
        for _dir in "test1", "test2", "test3":
            test_tempdir.joinpath(_dir).mkdir(parents=True, exist_ok=True)

            for _file in range(1, 4):
                test_tempdir.joinpath(_dir, str(_file)).touch()

        ret = salt_call_cli.run("state.apply", sls_name)
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True

        # Check to make sure the directories and files were created
        for _dir in "test1", "test2", "test3":
            assert test_tempdir.joinpath(_dir).is_dir()
            for _file in range(1, 4):
                assert test_tempdir.joinpath(_dir, str(_file)).is_file()


@pytest.mark.skip_on_windows
@pytest.mark.usefixtures("pillar_tree")
def test_recurse_keep_symlinks_in_fileserver_root(
    salt_master,
    salt_call_cli,
    tmp_path,
    salt_minion,
):
    target_path = tmp_path / "test_recurse"

    sls_name = "file_recurse_test"
    sls_contents = """
    tmp_recurse_dir_dir:
      file.recurse:
        - name: {target_path}
        - source: salt://tmp_dir/
        - keep_symlinks: True
    """.format(
        target_path=target_path
    )

    test_tempdir = salt_master.state_tree.base.write_path / "tmp_dir"
    test_tempdir.mkdir(parents=True, exist_ok=True)
    sls_tempfile = salt_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )

    with sls_tempfile:
        for _dir in "test1", "test2", "test3":
            test_tempdir.joinpath(_dir).mkdir(parents=True, exist_ok=True)

            for _file in range(1, 4):
                test_tempdir.joinpath(_dir, str(_file)).touch()

        with pytest.helpers.change_cwd(str(test_tempdir)):
            pathlib.Path("test").symlink_to("test3", target_is_directory=True)

        ret = salt_call_cli.run("state.apply", sls_name)
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True

        # Check to make sure the directories and files were created
        for _dir in "test1", "test2", "test3":
            assert target_path.joinpath(_dir).is_dir()
            for _file in range(1, 4):
                assert target_path.joinpath(_dir, str(_file)).is_file()

        assert target_path.joinpath("test").is_symlink()


@pytest.mark.skip_on_windows
@pytest.mark.usefixtures("pillar_tree")
def test_recurse_keep_symlinks_outside_fileserver_root(
    salt_secondary_minion,
    salt_secondary_master,
    salt_cli_secondary_wrapper,
    tmp_path,
):
    target_path = tmp_path / "test_recurse"

    sls_name = "file_recurse_test"
    sls_contents = """
    tmp_recurse_dir_dir:
      file.recurse:
        - name: {target_path}
        - source: salt://tmp_dir/
        - keep_symlinks: True
    """.format(
        target_path=target_path
    )

    test_tempdir = salt_secondary_master.state_tree.base.write_path / "tmp_dir"
    test_tempdir.mkdir(parents=True, exist_ok=True)
    sls_tempfile = salt_secondary_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )

    with sls_tempfile:
        for _dir in "test1", "test2", "test3":
            test_tempdir.joinpath(_dir).mkdir(parents=True, exist_ok=True)

            for _file in range(1, 4):
                test_tempdir.joinpath(_dir, str(_file)).touch()

        with pytest.helpers.change_cwd(str(test_tempdir)):
            pathlib.Path("/tmp/test_recurse_outside").mkdir(parents=True, exist_ok=True)
            pathlib.Path("test4").symlink_to(
                "/tmp/test_recurse_outside", target_is_directory=True
            )

        ret = salt_cli_secondary_wrapper("state.apply", sls_name)
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True

        # Check to make sure the directories and files were created
        for _dir in "test1", "test2", "test3":
            assert target_path.joinpath(_dir).is_dir()
            for _file in range(1, 4):
                assert target_path.joinpath(_dir, str(_file)).is_file()

        assert target_path.joinpath("test4").is_symlink()


@pytest.mark.usefixtures("pillar_tree")
def test_issue_62117(
    salt_master,
    salt_call_cli,
    tmp_path,
    salt_minion,
):
    name = "test_jinja/issue-62117"

    yaml_contents = "{%- set grains = grains.get('os', '') %}"

    jinja_contents = '{%- import_yaml "./issue-62117.yaml" as grains %}'

    if salt.utils.platform.is_windows():
        cmd = "cd"
    else:
        cmd = "pwd"
    sls_contents = f"""
    {{%- from "./issue-62117.jinja" import grains with context %}}

    test_jinja/issue-62117/cmd.run:
      cmd.run:
        - name: {cmd}
    """

    yaml_tempfile = salt_master.state_tree.base.temp_file(f"{name}.yaml", yaml_contents)

    jinja_tempfile = salt_master.state_tree.base.temp_file(
        f"{name}.jinja", jinja_contents
    )

    sls_tempfile = salt_master.state_tree.base.temp_file(f"{name}.sls", sls_contents)

    with yaml_tempfile, jinja_tempfile, sls_tempfile:
        ret = salt_call_cli.run("--local", "state.apply", name.replace("/", "."))
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["result"] is True


@pytest.mark.usefixtures("pillar_tree")
def test_issue_62611(
    salt_master,
    salt_call_cli,
    tmp_path,
    salt_minion,
):
    name = "test_jinja/issue-62611"

    jinja_contents = '{% set myvar = "MOOP" -%}'

    jinja_file = salt_master.state_tree.base.write_path / f"{name}.jinja"
    if salt.utils.platform.is_windows():
        jinja_file = str(jinja_file).replace("\\", "\\\\")

    sls_contents = """
    {%- from "REPLACEME" import myvar with context %}

    test_jinja/issue-62611/cmd.run:
      cmd.run:
        - name: echo MEEP {{ myvar }}
    """.replace(
        "REPLACEME", str(jinja_file)
    )

    jinja_tempfile = salt_master.state_tree.base.temp_file(
        f"{name}.jinja", jinja_contents
    )

    sls_tempfile = salt_master.state_tree.base.temp_file(f"{name}.sls", sls_contents)

    with jinja_tempfile, sls_tempfile:
        ret = salt_call_cli.run("--local", "state.apply", name.replace("/", "."))
        assert ret.returncode == 0
        assert ret.data
        state_run = next(iter(ret.data.values()))
        assert state_run["name"] == "echo MEEP MOOP"
        assert state_run["result"] is True


def test_contents_file(salt_master, salt_call_cli, tmp_path):
    """
    test calling file.managed multiple times
    with salt-call
    """
    target_path = tmp_path / "add-contents-file.txt"
    sls_name = "file-contents"
    sls_contents = """
    add_contents_file_sls:
      file.managed:
        - name: {}
        - contents: 1234
    """.format(
        target_path
    )
    sls_tempfile = salt_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )
    with sls_tempfile:
        for i in range(1, 4):
            ret = salt_call_cli.run("state.sls", sls_name)
            assert ret.returncode == 0
            assert ret.data
            state_run = next(iter(ret.data.values()))
            assert state_run["result"] is True
            # Check to make sure the file was created
            assert target_path.is_file()


@pytest.mark.skip_on_windows
def test_directory_recurse(salt_master, salt_call_cli, tmp_path, grains):
    """
    Test modifying ownership of symlink without affecting the link target's
    permissions.
    """
    target_dir = tmp_path / "test-dir"
    target_dir.mkdir()

    target_file = target_dir / "test-file"
    target_file.write_text("this is a test file")
    file_perms = target_file.stat().st_mode

    target_link = target_dir / "test-link"
    target_link.symlink_to(target_file)

    # Change the ownership of the sybolic link to 'nobody'
    ret = subprocess.run(["chown", "-h", "nobody", str(target_link)], check=False)
    assert ret.returncode == 0

    if grains["os"] != "VMware Photon OS":
        file_perms |= stat.S_IROTH

    # The permissions of the file should be 644.
    assert target_file.stat().st_mode == file_perms

    sls_name = "test"
    sls_contents = f"""
    {target_dir}:
      file.directory:
        - user: root
        - recurse:
          - user
    """
    sls_tempfile = salt_master.state_tree.base.temp_file(
        f"{sls_name}.sls", sls_contents
    )
    with sls_tempfile:
        ret = salt_call_cli.run("state.sls", sls_name)
        key = f"file_|-{target_dir}_|-{target_dir}_|-directory"
        assert key in ret.json
        result = ret.json[key]
        assert "changes" in result and result["changes"]

    # Permissions of file should not have changed.
    assert target_file.stat().st_mode == file_perms



"""
Read pillar data from a Redis backend
=====================================

.. versionadded:: 2014.7.0

:depends:   - redis Python module (on master)

Salt Master Redis Configuration
===============================

The module shares the same base Redis connection variables as
:py:mod:`salt.returners.redis_return`. These variables go in your master
config file.

* ``redis.db`` - The Redis database to use. Defaults to ``0``.
* ``redis.host`` - The Redis host to connect to. Defaults to ``'salt'``.
* ``redis.port`` - The port that the Redis database is listening on. Defaults
  to ``6379``.
* ``redis.password`` - The password for authenticating with Redis. Only
  required if you are using master auth. Defaults to ``None``.

Configuring the Redis ext_pillar
================================

    .. code-block:: yaml

        ext_pillar:
          - redis: {function: key_value}

"""

import salt.utils.json

__virtualname__ = "redis"


def __virtual__():
    """
    Only load if the redis module is in __salt__
    """
    if "redis.get_key" in __salt__:
        return __virtualname__
    return False


def ext_pillar(minion_id, pillar, function, **kwargs):
    """
    Grabs external pillar data based on configured function
    """
    if function.startswith("_") or function not in globals():
        return {}
    # Call specified function to pull redis data
    return globals()[function](minion_id, pillar, **kwargs)


def key_value(minion_id, pillar, pillar_key="redis_pillar"):  # pylint: disable=W0613
    """
    Looks for key in redis matching minion_id, returns a structure based on the
    data type of the redis key. String for string type, dict for hash type and
    lists for lists, sets and sorted sets.

    pillar_key
        Pillar key to return data into
    """
    # Identify key type and process as needed based on that type
    key_type = __salt__["redis.key_type"](minion_id)
    if key_type == "string":
        return {pillar_key: __salt__["redis.get_key"](minion_id)}
    elif key_type == "hash":
        return {pillar_key: __salt__["redis.hgetall"](minion_id)}
    elif key_type == "list":
        list_size = __salt__["redis.llen"](minion_id)
        if not list_size:
            return {}
        return {pillar_key: __salt__["redis.lrange"](minion_id, 0, list_size - 1)}
    elif key_type == "set":
        return {pillar_key: __salt__["redis.smembers"](minion_id)}
    elif key_type == "zset":
        set_size = __salt__["redis.zcard"](minion_id)
        if not set_size:
            return {}
        return {pillar_key: __salt__["redis.zrange"](minion_id, 0, set_size - 1)}
    # Return nothing for unhandled types
    return {}


def key_json(minion_id, pillar, pillar_key=None):  # pylint: disable=W0613
    """
    Pulls a string from redis and deserializes it from json. Deserialized
    dictionary data loaded directly into top level if pillar_key is not set.

    pillar_key
        Pillar key to return data into
    """
    key_data = __salt__["redis.get_key"](minion_id)
    # Return nothing for non-existent keys
    if not key_data:
        return {}

    data = salt.utils.json.loads(key_data)
    # Return as requested
    if isinstance(data, dict) and not pillar_key:
        return data
    elif not pillar_key:
        return {"redis_pillar": data}
    else:
        return {pillar_key: data}



import logging
import os

import pytest

import salt.states.file as filestate
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


pytestmark = pytest.mark.skipif(
    pytest.mark.skip_unless_on_linux, reason="Only run on Linux"
)


@pytest.fixture
def configure_loader_modules():
    return {
        filestate: {
            "__env__": "base",
            "__salt__": {"file.manage_file": False},
            "__opts__": {"test": False, "cachedir": ""},
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {},
        }
    }


def test_selinux_change():
    file_name = "/tmp/some-test-file"
    check_perms_result = [
        {
            "comment": f"The file {file_name} is set to be changed",
            "changes": {
                "selinux": {
                    "New": "User: unconfined_u Type: lost_found_t",
                    "Old": "User: system_u Type: user_tmp_t",
                }
            },
            "name": file_name,
            "result": True,
        },
        {"luser": "root", "lmode": "0644", "lgroup": "root"},
    ]

    with patch.object(os.path, "exists", MagicMock(return_value=True)):
        with patch.dict(
            filestate.__salt__,
            {
                "file.source_list": MagicMock(return_value=[file_name, None]),
                "file.check_perms": MagicMock(return_value=check_perms_result),
            },
        ):
            ret = filestate.managed(
                file_name,
                selinux={"seuser": "unconfined_u", "setype": "user_tmp_t"},
            )
            assert ret["result"]



import re

import pytest

import salt.defaults.exitcodes
import salt.utils.files
import salt.utils.platform
import salt.utils.pycrypto
import salt.utils.yaml

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def salt_run_cli(salt_master):
    """
    Override salt_run_cli fixture to provide an increased default_timeout to the calls
    """
    return salt_master.salt_run_cli(timeout=120)


def test_in_docs(salt_run_cli):
    """
    test the salt-run docs system
    """
    ret = salt_run_cli.run("-d")
    assert "jobs.active:" in ret.stdout
    assert "jobs.list_jobs:" in ret.stdout
    assert "jobs.lookup_jid:" in ret.stdout
    assert "manage.down:" in ret.stdout
    assert "manage.up:" in ret.stdout
    assert "network.wol:" in ret.stdout
    assert "network.wollist:" in ret.stdout


def test_not_in_docs(salt_run_cli):
    """
    test the salt-run docs system
    """
    ret = salt_run_cli.run("-d")
    assert "jobs.SaltException:" not in ret.stdout


def test_salt_documentation_too_many_arguments(salt_run_cli):
    """
    Test to see if passing additional arguments shows an error
    """
    ret = salt_run_cli.run("-d", "virt.list", "foo")
    assert ret.returncode != 0
    assert "You can only get documentation for one method at one time" in ret.stderr


def test_exit_status_unknown_argument(salt_run_cli):
    """
    Ensure correct exit status when an unknown argument is passed to salt-run.
    """
    ret = salt_run_cli.run("--unknown-argument")
    assert ret.returncode == salt.defaults.exitcodes.EX_USAGE, ret
    assert "Usage" in ret.stderr
    assert "no such option: --unknown-argument" in ret.stderr


def test_exit_status_correct_usage(salt_run_cli):
    """
    Ensure correct exit status when salt-run starts correctly.
    """
    ret = salt_run_cli.run("test.arg", "arg1", kwarg1="kwarg1")
    assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret


@pytest.mark.skip_if_not_root
@pytest.mark.parametrize("flag", ["--auth", "--eauth", "--external-auth", "-a"])
@pytest.mark.skip_on_windows(reason="PAM is not supported on Windows")
def test_salt_run_with_eauth_all_args(salt_run_cli, salt_eauth_account, flag):
    """
    test salt-run with eauth
    tests all eauth args
    """
    ret = salt_run_cli.run(
        flag,
        "pam",
        "--username",
        salt_eauth_account.username,
        "--password",
        salt_eauth_account.password,
        "test.arg",
        "arg",
        kwarg="kwarg1",
        _timeout=240,
    )
    assert ret.returncode == 0, ret
    assert ret.data, ret
    expected = {"args": ["arg"], "kwargs": {"kwarg": "kwarg1"}}
    assert ret.data == expected, ret


@pytest.mark.skip_if_not_root
@pytest.mark.skip_on_windows(reason="PAM is not supported on Windows")
def test_salt_run_with_eauth_bad_passwd(salt_run_cli, salt_eauth_account):
    """
    test salt-run with eauth and bad password
    """
    ret = salt_run_cli.run(
        "-a",
        "pam",
        "--username",
        salt_eauth_account.username,
        "--password",
        "wrongpassword",
        "test.arg",
        "arg",
        kwarg="kwarg1",
    )
    assert (
        ret.stdout
        == 'Authentication failure of type "eauth" occurred for user {}.'.format(
            salt_eauth_account.username
        )
    )


@pytest.mark.skip_if_not_root
def test_salt_run_with_wrong_eauth(salt_run_cli, salt_eauth_account):
    """
    test salt-run with wrong eauth parameter
    """
    ret = salt_run_cli.run(
        "-a",
        "wrongeauth",
        "--username",
        salt_eauth_account.username,
        "--password",
        salt_eauth_account.password,
        "test.arg",
        "arg",
        kwarg="kwarg1",
    )
    assert ret.returncode == 0, ret
    assert re.search(
        r"^The specified external authentication system \"wrongeauth\" is not"
        r" available\nAvailable eauth types: auto, .*",
        ret.stdout.replace("\r\n", "\n"),
    )



import pytest
from saltfactories.utils import random_string

import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from tests.support.case import ModuleCase

if not salt.utils.platform.is_windows():
    import grp


@pytest.mark.skip_if_not_root
@pytest.mark.destructive_test
@pytest.mark.windows_whitelisted
@pytest.mark.skip_unless_on_platforms(linux=True, windows=True)
class GroupModuleTest(ModuleCase):
    """
    Validate the linux group system module
    """

    def setUp(self):
        """
        Get current settings
        """
        super().setUp()
        self._user = random_string("tg-", uppercase=False)
        self._user1 = random_string("tg-", uppercase=False)
        self._no_user = random_string("tg-", uppercase=False)
        self._group = random_string("tg-", uppercase=False)
        self._no_group = random_string("tg-", uppercase=False)
        _gid = _new_gid = None
        if not salt.utils.platform.is_windows():
            _gid = 64989
            _new_gid = 64998
        self._gid = _gid
        self._new_gid = _new_gid

    def tearDown(self):
        """
        Reset to original settings
        """
        self.run_function("user.delete", [self._user])
        self.run_function("user.delete", [self._user1])
        self.run_function("group.delete", [self._group])

    def __get_system_group_gid_range(self):
        """
        Returns (SYS_GID_MIN, SYS_GID_MAX)
        """
        try:
            login_defs = {}
            with salt.utils.files.fopen("/etc/login.defs") as defs_fd:
                for line in defs_fd:
                    line = salt.utils.stringutils.to_unicode(line).strip()
                    if line.startswith("#"):
                        continue
                    try:
                        key, val = line.split()
                    except ValueError:
                        pass
                    else:
                        login_defs[key] = val
        except OSError:
            login_defs = {"SYS_GID_MIN": 101, "SYS_GID_MAX": 999}

        gid_min = login_defs.get("SYS_GID_MIN", 101)
        gid_max = login_defs.get(
            "SYS_GID_MAX", int(login_defs.get("GID_MIN", 1000)) - 1
        )

        return int(gid_min), int(gid_max)

    def __get_free_system_gid(self):
        """
        Find a free system gid
        """

        gid_min, gid_max = self.__get_system_group_gid_range()

        busy_gids = [x.gr_gid for x in grp.getgrall() if gid_min <= x.gr_gid <= gid_max]

        # find free system gid
        for gid in range(gid_min, gid_max + 1):
            if gid not in busy_gids:
                return gid

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_add(self):
        """
        Test the add group function
        """
        # add a new group
        self.assertTrue(self.run_function("group.add", [self._group], gid=self._gid))
        group_info = self.run_function("group.info", [self._group])
        self.assertEqual(group_info["gid"], self._gid)
        self.assertEqual(group_info["name"], self._group)
        # try adding the group again
        self.assertFalse(self.run_function("group.add", [self._group], gid=self._gid))

    @pytest.mark.destructive_test
    @pytest.mark.skip_on_windows(reason="Skip on Windows")
    @pytest.mark.slow_test
    def test_add_system_group(self):
        """
        Test the add group function with system=True
        """

        gid_min, gid_max = self.__get_system_group_gid_range()

        # add a new system group
        self.assertTrue(self.run_function("group.add", [self._group, None, True]))
        group_info = self.run_function("group.info", [self._group])
        self.assertEqual(group_info["name"], self._group)
        self.assertTrue(gid_min <= group_info["gid"] <= gid_max)
        # try adding the group again
        self.assertFalse(self.run_function("group.add", [self._group]))

    @pytest.mark.destructive_test
    @pytest.mark.skip_on_windows(reason="Skip on Windows")
    @pytest.mark.slow_test
    def test_add_system_group_gid(self):
        """
        Test the add group function with system=True and a specific gid
        """

        gid = self.__get_free_system_gid()

        # add a new system group
        self.assertTrue(self.run_function("group.add", [self._group, gid, True]))
        group_info = self.run_function("group.info", [self._group])
        self.assertEqual(group_info["name"], self._group)
        self.assertEqual(group_info["gid"], gid)
        # try adding the group again
        self.assertFalse(self.run_function("group.add", [self._group, gid]))

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_delete(self):
        """
        Test the delete group function
        """
        self.assertTrue(self.run_function("group.add", [self._group]))

        # correct functionality
        self.assertTrue(self.run_function("group.delete", [self._group]))

        # group does not exist
        self.assertFalse(self.run_function("group.delete", [self._no_group]))

    @pytest.mark.slow_test
    def test_info(self):
        """
        Test the info group function
        """
        self.run_function("group.add", [self._group], gid=self._gid)
        self.run_function("user.add", [self._user])
        self.run_function("group.adduser", [self._group, self._user])
        group_info = self.run_function("group.info", [self._group])

        self.assertEqual(group_info["name"], self._group)
        self.assertEqual(group_info["gid"], self._gid)
        self.assertIn(self._user, str(group_info["members"]))

    @pytest.mark.skip_on_windows(reason="gid test skipped on windows")
    @pytest.mark.slow_test
    def test_chgid(self):
        """
        Test the change gid function
        """
        self.run_function("group.add", [self._group], gid=self._gid)
        self.assertTrue(self.run_function("group.chgid", [self._group, self._new_gid]))
        group_info = self.run_function("group.info", [self._group])
        self.assertEqual(group_info["gid"], self._new_gid)

    @pytest.mark.slow_test
    def test_adduser(self):
        """
        Test the add user to group function
        """
        self.run_function("group.add", [self._group], gid=self._gid)
        self.run_function("user.add", [self._user])
        self.assertTrue(self.run_function("group.adduser", [self._group, self._user]))
        group_info = self.run_function("group.info", [self._group])
        self.assertIn(self._user, str(group_info["members"]))
        # try add a non existing user
        self.assertFalse(
            self.run_function("group.adduser", [self._group, self._no_user])
        )
        # try add a user to non existing group
        self.assertFalse(
            self.run_function("group.adduser", [self._no_group, self._user])
        )
        # try add a non existing user to a non existing group
        self.assertFalse(
            self.run_function("group.adduser", [self._no_group, self._no_user])
        )

    @pytest.mark.slow_test
    def test_deluser(self):
        """
        Test the delete user from group function
        """
        self.run_function("group.add", [self._group], gid=self._gid)
        self.run_function("user.add", [self._user])
        self.run_function("group.adduser", [self._group, self._user])
        self.assertTrue(self.run_function("group.deluser", [self._group, self._user]))
        group_info = self.run_function("group.info", [self._group])
        self.assertNotIn(self._user, str(group_info["members"]))

    @pytest.mark.slow_test
    def test_members(self):
        """
        Test the members function
        """
        self.run_function("group.add", [self._group], gid=self._gid)
        self.run_function("user.add", [self._user])
        self.run_function("user.add", [self._user1])
        m = f"{self._user},{self._user1}"
        ret = self.run_function("group.members", [self._group, m])
        self.assertTrue(ret)
        group_info = self.run_function("group.info", [self._group])
        self.assertIn(self._user, str(group_info["members"]))
        self.assertIn(self._user1, str(group_info["members"]))

    @pytest.mark.slow_test
    def test_getent(self):
        """
        Test the getent function
        """
        self.run_function("group.add", [self._group], gid=self._gid)
        self.run_function("user.add", [self._user])
        self.run_function("group.adduser", [self._group, self._user])
        ginfo = self.run_function("user.getent")
        self.assertIn(self._group, str(ginfo))
        self.assertIn(self._user, str(ginfo))
        self.assertNotIn(self._no_group, str(ginfo))
        self.assertNotIn(self._no_user, str(ginfo))



"""
Unit tests for the LGPO module
"""

import platform

import pytest
from packaging import version

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


def _parse_platform_version():
    if platform.system() != "Windows":
        # Test only applies to windows but pytestmark is not getting
        # evaluated soon enough
        return True
    return version.parse(platform.version()) < version.parse("10.0.16299")


@pytest.fixture(scope="module")
def lgpo(modules):
    return modules.lgpo


# The "Allow Online Tips" policy only became available in version 10.0.16299
@pytest.mark.skipif(
    _parse_platform_version(),
    reason="Policy only available on 10.0.16299 or later",
)
def test_62058_whitespace(lgpo):
    result = lgpo.get_policy_info("Allow Online Tips", "machine")
    for element in result["policy_elements"]:
        if "element_aliases" in element:
            assert (
                "Allow Settings to retrieve online tips." in element["element_aliases"]
            )
            return
    assert False


def test_61859(lgpo):
    expected = (
        'ADMX policy name/id "Pol_CipherSuiteOrder" is used in multiple ADMX files.\n'
        "Try one of the following names:\n"
        " - Lanman Server\\Network\\Cipher suite order\n"
        " - Lanman Workstation\\Network\\Cipher suite order"
    )
    result = lgpo.get_policy_info(
        policy_name="Pol_CipherSuiteOrder",
        policy_class="machine",
    )
    assert result["message"] == expected



import ctypes
import logging
import multiprocessing
import os
import pathlib
import shutil
import stat
import time
from pathlib import Path

import pytest
from saltfactories.utils import random_string

import salt.channel.client
import salt.channel.server
import salt.config
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.master
import salt.utils.platform
import salt.utils.process
import salt.utils.stringutils
from tests.conftest import FIPS_TESTRUN

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_on_spawning_platform(
        reason="These tests are currently broken on spawning platforms. Need to be rewritten.",
    ),
    pytest.mark.timeout_unless_on_windows(120),
]


@pytest.fixture
def channel_minion_id():
    return random_string("Tha-Minion-")


@pytest.fixture
def root_dir(tmp_path):
    if salt.utils.platform.is_darwin():
        # To avoid 'OSError: AF_UNIX path too long'
        _root_dir = pathlib.Path("/tmp").resolve() / tmp_path.name
        try:
            yield _root_dir
        finally:
            shutil.rmtree(str(_root_dir), ignore_errors=True)
    else:
        yield tmp_path


def transport_ids(value):
    return f"transport({value})"


@pytest.fixture(params=["tcp", "zeromq"], ids=transport_ids)
def transport(request):
    return request.param


@pytest.fixture
def master_config(master_opts, transport):
    master_opts.update(
        transport=transport,
        id="master",
        interface="127.0.0.1",
        fips_mode=FIPS_TESTRUN,
        publish_signing_algorithm=(
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    )
    salt.crypt.gen_keys(master_opts["pki_dir"], "master", 4096)
    yield master_opts


@pytest.fixture
def minion_config(minion_opts, master_config, channel_minion_id):
    minion_opts.update(
        transport=master_config["transport"],
        root_dir=master_config["root_dir"],
        id=channel_minion_id,
        cachedir=master_config["cachedir"],
        sock_dir=master_config["sock_dir"],
        ret_port=master_config["ret_port"],
        interface="127.0.0.1",
        pki_dir=os.path.join(master_config["root_dir"], "pki_minion"),
        master_port=master_config["ret_port"],
        master_ip="127.0.0.1",
        master_uri="tcp://127.0.0.1:{}".format(master_config["ret_port"]),
        fips_mode=FIPS_TESTRUN,
        encryption_algorithm="OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        signing_algorithm="PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    )
    pathlib.Path(minion_opts["pki_dir"]).mkdir(exist_ok=True)
    pathlib.Path(master_config["pki_dir"]).mkdir(exist_ok=True)
    salt.crypt.gen_keys(minion_opts["pki_dir"], "minion", 4096)
    minion_pub = os.path.join(minion_opts["pki_dir"], "minion.pub")
    pub_on_master = os.path.join(master_config["pki_dir"], "minions", channel_minion_id)
    shutil.copyfile(minion_pub, pub_on_master)
    return minion_opts


@pytest.fixture
def process_manager():
    process_manager = salt.utils.process.ProcessManager()
    try:
        yield process_manager
    finally:
        process_manager.terminate()


@pytest.fixture
def master_secrets():
    salt.master.SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "serial": multiprocessing.Value(
            ctypes.c_longlong, lock=False  # We'll use the lock from 'secret'
        ),
    }
    yield
    salt.master.SMaster.secrets.pop("aes")


@salt.ext.tornado.gen.coroutine
def _connect_and_publish(
    io_loop, channel_minion_id, channel, server, received, timeout=60
):
    log.info("TEST - BEFORE CHANNEL CONNECT")
    yield channel.connect()
    log.info("TEST - AFTER CHANNEL CONNECT")

    def cb(payload):
        log.info("TEST - PUB SERVER MSG %r", payload)
        received.append(payload)
        io_loop.stop()

    channel.on_recv(cb)
    server.publish({"tgt_type": "glob", "tgt": [channel_minion_id], "WTF": "SON"})
    start = time.time()
    while time.time() - start < timeout:
        yield salt.ext.tornado.gen.sleep(1)
    io_loop.stop()


def test_pub_server_channel(
    io_loop,
    channel_minion_id,
    master_config,
    minion_config,
    process_manager,
    master_secrets,
):
    server_channel = salt.channel.server.PubServerChannel.factory(
        master_config,
    )
    server_channel.pre_fork(process_manager)
    req_server_channel = salt.channel.server.ReqServerChannel.factory(master_config)
    req_server_channel.pre_fork(process_manager)

    def handle_payload(payload):
        log.info("TEST - Req Server handle payload %r", payload)

    req_server_channel.post_fork(handle_payload, io_loop=io_loop)

    if master_config["transport"] == "zeromq":
        time.sleep(1)
        attempts = 5
        while True:
            try:
                p = Path(str(master_config["sock_dir"])) / "workers.ipc"
                mode = os.lstat(p).st_mode
                assert bool(os.lstat(p).st_mode & stat.S_IRUSR)
                assert not bool(os.lstat(p).st_mode & stat.S_IRGRP)
                assert not bool(os.lstat(p).st_mode & stat.S_IROTH)
                break
            except FileNotFoundError as exc:
                if not attempts:
                    raise exc from None
                attempts -= 1
                time.sleep(2.5)

    pub_channel = salt.channel.client.AsyncPubChannel.factory(minion_config)
    received = []

    try:
        io_loop.add_callback(
            _connect_and_publish,
            io_loop,
            channel_minion_id,
            pub_channel,
            server_channel,
            received,
        )
        io_loop.start()
        assert len(received) == 1
    finally:
        server_channel.close()
        req_server_channel.close()
        pub_channel.close()



"""
Display profiling data in a table format
========================================

Show profile data for returners that would normally show a highstate output.

CLI Example:

.. code-block:: bash

    salt '*' state.apply something --out=profile

Attempt to output the returns of state.sls and state.highstate as a table of
names, modules and durations that looks somewhat like the following::

    name                mod.fun                duration (ms)
    --------------------------------------------------------
    I-fail-unless-stmt  other.function               -1.0000
    old-minion-config   grains.list_present           1.1200
    salt-data           group.present                48.3800
    /etc/salt/minion    file.managed                 63.1450


To get the above appearance, use settings something like these::

    out.table.separate_rows: False
    out.table.justify: left
    out.table.delim: '  '
    out.table.prefix: ''
    out.table.suffix: ''
"""

import salt.output.table_out as table_out

__virtualname__ = "profile"


def __virtual__():
    return True


def _find_durations(data, name_max=60):
    ret = []
    ml = len("duration (ms)")
    for host in data:
        for sid in data[host]:
            try:
                dat = data[host][sid]
            except TypeError:
                dat = {}

            if not isinstance(dat, dict):
                dat = {"name": str(sid)}

            ts = str(sid).split("_|-")
            mod = ts[0]
            fun = ts[-1]
            name = dat.get("name", dat.get("__id__"))
            dur = float(dat.get("duration", -1))

            if name is None:
                name = "<>"
            if len(name) > name_max:
                name = name[0 : name_max - 3] + "..."

            l = len(f"{dur:0.4f}")
            if l > ml:
                ml = l

            ret.append([dur, name, f"{mod}.{fun}"])

    for row in ret:
        row[0] = "{0:{w}.4f}".format(row[0], w=ml)
    return [x[1:] + x[0:1] for x in sorted(ret)]


def output(data, **kwargs):
    """
    Display the profiling data in a table format.
    """

    rows = _find_durations(data)

    kwargs["opts"] = __opts__
    kwargs["rows_key"] = "rows"
    kwargs["labels_key"] = "labels"

    to_show = {"labels": ["name", "mod.fun", "duration (ms)"], "rows": rows}

    return table_out.output(to_show, **kwargs)



"""
PeeringDB Module
================

.. versionadded:: 2019.2.0

Execution module for the basic interaction with the
`PeeringDB <https://www.peeringdb.com/>`_ API.

While for GET operations (the functions prefixed by ``get_``) the credentials
are optional, there are some specific details that are visible only to
authenticated users. Moreover, the credentials are required when adding or
updating information. That means, the module can equally work out of the box
without any further configuration with the limitations imposed by the PeeringDB
API.

For complete API documentation, please refer to https://www.peeringdb.com/apidocs/.

Configuration (in the opts or Pillar):

.. code-block:: yaml

    peeringdb:
      username: salt
      password: 5@1t
"""

import logging

# Import salt modules
import salt.utils.http
from salt.utils.args import clean_kwargs

log = logging.getLogger(__name__)


__virtualname__ = "peeringdb"
__proxyenabled__ = ["*"]

PEERINGDB_URL = "https://www.peeringdb.com/api"


def __virtual__():
    return __virtualname__


def _get_auth(username=None, password=None):
    peeringdb_cfg = __salt__["config.merge"]("peeringdb", default={})
    if not username:
        username = peeringdb_cfg.get("username", username)
    if not password:
        password = peeringdb_cfg.get("password", password)
    return username, password


def _build_url(endpoint, id=None):
    if id:
        return f"{PEERINGDB_URL}/{endpoint}/{id}"
    return f"{PEERINGDB_URL}/{endpoint}"


def _get_endpoint(endpoint, id=None, **kwargs):
    username, password = _get_auth(
        kwargs.pop("username", None), kwargs.pop("password", None)
    )
    kwargs = clean_kwargs(**kwargs)
    url = _build_url(endpoint, id=id)
    ret = {"comment": "", "result": True, "out": None}
    res = salt.utils.http.query(
        url,
        method="GET",
        decode=True,
        username=username,
        password=password,
        params=kwargs,
    )
    if "error" in res:
        ret.update({"result": False, "comment": res["error"]})
        return ret
    ret["out"] = res["dict"]["data"]
    return ret


def get_net(**kwargs):
    """
    Return the details of a network identified using the search filters
    specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible networks registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/net/net_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_net id=4224
        salt '*' peeringdb.get_net asn=13335
        salt '*' peeringdb.get_net city='Salt Lake City'
        salt '*' peeringdb.get_net name__startswith=GTT
    """
    return _get_endpoint("net", **kwargs)


def get_fac(**kwargs):
    """
    Return the details of the facility identified using the search
    filters specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible facilities registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/netfac/netfac_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_fac id=1774
        salt '*' peeringdb.get_fac state=UT
    """
    return _get_endpoint("fac", **kwargs)


def get_ix(**kwargs):
    """
    Return the details of an IX (Internet Exchange) using the search filters
    specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible IXs registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/ix/ix_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_ix id=1
        salt '*' peeringdb.get_ix city='Milwaukee'
    """
    return _get_endpoint("ix", **kwargs)


def get_ixfac(**kwargs):
    """
    Return the details of an IX (Internet Exchange) facility using the search
    filters specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible IX facilities registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/ixfac/ixfac_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_ixfac id=1
        salt '*' peeringdb.get_ixfac city='Milwaukee'
    """
    return _get_endpoint("ixfac", **kwargs)


def get_ixlan(**kwargs):
    """
    Return the details of an IX (Internet Exchange) together with the networks
    available in this location (and their details), using the search filters
    specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible IX LAN facilities registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/ixlan/ixlan_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_ixlan id=780
        salt '*' peeringdb.get_ixlan city='Milwaukee'
    """
    return _get_endpoint("ixlan", **kwargs)


def get_ixpfx(**kwargs):
    """
    Return the details of an IX (Internet Exchange) together with the PeeringDB
    IDs of the networks available in this location, using the search filters
    specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible IX LAN facilities registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/ixpfx/ixpfx_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_ixpfx id=780
        salt '*' peeringdb.get_ixpfx city='Milwaukee'
    """
    return _get_endpoint("ixpfx", **kwargs)


def get_netfac(**kwargs):
    """
    Return the list of facilities used by a particular network, given the ``id``
    or other filters specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible network facilities registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/netfac/netfac_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_netfac id=780
        salt '*' peeringdb.get_netfac city='Milwaukee'
    """
    return _get_endpoint("netfac", **kwargs)


def get_netixlan(**kwargs):
    """
    Return the IP addresses used by a particular network at all the IXs where it
    is available. The network is selected either via the ``id`` argument or the
    other filters specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible IP addresses, of all networks, at all IXs, registered in
        PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/netixlan/netixlan_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_netixlan asn=13335
        salt '*' peeringdb.get_netixlan ipaddr4=185.1.114.25
    """
    return _get_endpoint("netixlan", **kwargs)


def get_org(**kwargs):
    """
    Return the details of an organisation together with the networks
    available in this location, using the search filters specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible organisations registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/org/org_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_org id=2
        salt '*' peeringdb.get_org city=Duesseldorf
    """
    return _get_endpoint("org", **kwargs)


def get_poc(**kwargs):
    """
    Return the details of a person of contact together using the search filters
    specified in the query.

    .. note::
        If no ``id`` or filter arguments are specified, it will return all the
        possible contacts registered in PeeringDB.

        The available filters are documented at:
        https://www.peeringdb.com/apidocs/#!/poc/poc_list

    CLI Example:

    .. code-block:: bash

        salt '*' peeringdb.get_poc id=6721
        salt '*' peeringdb.get_poc email__contains='@cloudflare.com'
    """
    return _get_endpoint("poc", **kwargs)
